# -*- coding: utf-8 -*-

from typing import Union, Dict, Callable

import brainpy.math as bm
from brainpy.connect import TwoEndConnector, All2All, One2One
from brainpy.dyn.base import NeuGroup, TwoEndConn
from brainpy.initialize import Initializer, init_param
from brainpy.integrators import odeint
from brainpy.types import Tensor

__all__ = [
  'AMPA',
  'GABAa',
]


class AMPA(TwoEndConn):
  r"""AMPA conductance-based synapse model.

  **Model Descriptions**

  AMPA receptor is an ionotropic receptor, which is an ion channel.
  When it is bound by neurotransmitters, it will immediately open the
  ion channel, causing the change of membrane potential of postsynaptic neurons.

  A classical model is to use the Markov process to model ion channel switch.
  Here :math:`g` represents the probability of channel opening, :math:`1-g`
  represents the probability of ion channel closing, and :math:`\alpha` and
  :math:`\beta` are the transition probability. Because neurotransmitters can
  open ion channels, the transfer probability from :math:`1-g` to :math:`g`
  is affected by the concentration of neurotransmitters. We denote the concentration
  of neurotransmitters as :math:`[T]` and get the following Markov process.

  .. image:: ../../../../_static/synapse_markov.png
      :align: center

  We obtained the following formula when describing the process by a differential equation.

  .. math::

      \frac{ds}{dt} =\alpha[T](1-g)-\beta g

  where :math:`\alpha [T]` denotes the transition probability from state :math:`(1-g)`
  to state :math:`(g)`; and :math:`\beta` represents the transition probability of
  the other direction. :math:`\alpha` is the binding constant. :math:`\beta` is the
  unbinding constant. :math:`[T]` is the neurotransmitter concentration, and
  has the duration of 0.5 ms.

  Moreover, the post-synaptic current on the post-synaptic neuron is formulated as

  .. math::

      I_{syn} = g_{max} g (V-E)

  where :math:`g_{max}` is the maximum conductance, and `E` is the reverse potential.

  **Model Examples**


  .. plot::
    :include-source: True

    >>> import brainpy as bp
    >>> import matplotlib.pyplot as plt
    >>>
    >>> neu1 = bp.dyn.HH(1)
    >>> neu2 = bp.dyn.HH(1)
    >>> syn1 = bp.dyn.AMPA(neu1, neu2, bp.connect.All2All())
    >>> net = bp.dyn.Network(pre=neu1, syn=syn1, post=neu2)
    >>>
    >>> runner = bp.dyn.DSRunner(net, inputs=[('pre.input', 5.)], monitors=['pre.V', 'post.V', 'syn.g'])
    >>> runner.run(150.)
    >>>
    >>> fig, gs = bp.visualize.get_figure(2, 1, 3, 8)
    >>> fig.add_subplot(gs[0, 0])
    >>> plt.plot(runner.mon.ts, runner.mon['pre.V'], label='pre-V')
    >>> plt.plot(runner.mon.ts, runner.mon['post.V'], label='post-V')
    >>> plt.legend()
    >>>
    >>> fig.add_subplot(gs[1, 0])
    >>> plt.plot(runner.mon.ts, runner.mon['syn.g'], label='g')
    >>> plt.legend()
    >>> plt.show()

  **Model Parameters**

  ============= ============== ======== ================================================
  **Parameter** **Init Value** **Unit** **Explanation**
  ------------- -------------- -------- ------------------------------------------------
  delay         0              ms       The decay length of the pre-synaptic spikes.
  g_max         .42            µmho(µS) Maximum conductance.
  E             0              mV       The reversal potential for the synaptic current.
  alpha         .98            \        Binding constant.
  beta          .18            \        Unbinding constant.
  T             .5             mM       Neurotransmitter concentration.
  T_duration    .5             ms       Duration of the neurotransmitter concentration.
  ============= ============== ======== ================================================


  **Model Variables**

  ================== ================== ==================================================
  **Member name**    **Initial values** **Explanation**
  ------------------ ------------------ --------------------------------------------------
  g                  0                  Synapse gating variable.
  pre_spike          False              The history of pre-synaptic neuron spikes.
  spike_arrival_time -1e7               The arrival time of the pre-synaptic neuron spike.
  ================== ================== ==================================================

  **References**

  .. [1] Vijayan S, Kopell N J. Thalamic model of awake alpha oscillations
         and implications for stimulus processing[J]. Proceedings of the
         National Academy of Sciences, 2012, 109(45): 18553-18558.
  """

  def __init__(
      self,
      pre: NeuGroup,
      post: NeuGroup,
      conn: Union[TwoEndConnector, Tensor, Dict[str, Tensor]],
      conn_type: str = 'dense',
      g_max: Union[float, Tensor, Initializer, Callable] = 0.42,
      E: float = 0.,
      alpha: float = 0.98,
      beta: float = 0.18,
      T: float = 0.5,
      T_duration: float = 0.5,
      delay_step: Union[int, Tensor, Initializer, Callable] = None,
      method: str = 'exp_auto',
      name: str = None
  ):
    super(AMPA, self).__init__(pre=pre, post=post, conn=conn, name=name)
    self.check_pre_attrs('spike')
    self.check_post_attrs('input', 'V')

    # parameters
    self.g_max = g_max
    self.E = E
    self.alpha = alpha
    self.beta = beta
    self.T = T
    self.T_duration = T_duration

    # connection
    self.conn_type = conn_type
    if conn_type not in ['sparse', 'dense']:
      raise ValueError(f'"conn_type" must be in "sparse" and "dense", but we got {conn_type}')
    if self.conn is None:
      raise ValueError(f'Must provide "conn" when initialize the model {self.name}')
    if isinstance(self.conn, One2One):
      self.g_max = init_param(g_max, (self.pre.num,), allow_none=False)
      self.weight_type = 'heter' if bm.size(self.g_max) != 1 else 'homo'
    elif isinstance(self.conn, All2All):
      self.g_max = init_param(g_max, (self.pre.num, self.post.num), allow_none=False)
      if bm.size(self.g_max) != 1:
        self.weight_type = 'heter'
        bm.fill_diagonal(self.g_max, 0.)
      else:
        self.weight_type = 'homo'
    else:
      if conn_type == 'sparse':
        self.pre_ids, self.post_ids = self.conn.require('pre_ids', 'post_ids')
        self.g_max = init_param(g_max, self.post_ids.shape, allow_none=False)
        self.weight_type = 'heter' if bm.size(self.g_max) != 1 else 'homo'
      elif conn_type == 'dense':
        self.g_max = init_param(g_max, (self.pre.num, self.post.num), allow_none=False)
        self.weight_type = 'heter' if bm.size(self.g_max) != 1 else 'homo'
        if self.weight_type == 'homo':
          self.conn_mat = self.conn.require('conn_mat')
      else:
        raise ValueError(f'Unknown connection type: {conn_type}')

    # variables
    self.g = bm.Variable(bm.zeros(self.pre.num))
    self.spike_arrival_time = bm.Variable(bm.ones(self.pre.num) * -1e7)
    self.delay_step = self.register_delay(self.pre.name + '.spike', delay_step, self.pre.spike)

    # functions
    self.integral = odeint(method=method, f=self.dg)

  def dg(self, g, t, TT):
    dg = self.alpha * TT * (1 - g) - self.beta * g
    return dg

  def update(self, _t, _dt):
    # delays
    if self.delay_step is None:
      pre_spike = self.pre.spike
    else:
      pre_spike = self.get_delay(self.pre.name + '.spike', self.delay_step)
      self.update_delay(self.pre.name + '.spike', self.pre.spike)

    # spike arrival time
    self.spike_arrival_time.value = bm.where(pre_spike, _t, self.spike_arrival_time)

    # post-synaptic values
    TT = ((_t - self.spike_arrival_time) < self.T_duration) * self.T
    self.g.value = self.integral(self.g, _t, TT, dt=_dt)
    if isinstance(self.conn, One2One):
      post_g = self.g_max * self.g
    elif isinstance(self.conn, All2All):
      if self.weight_type == 'homo':
        post_g = self.g.sum() * self.g_max
        if not self.conn.include_self:
          post_g = post_g - self.g
      else:
        post_g = self.g @ self.g_max
    else:
      if self.conn_type == 'sparse':
        post_g = bm.pre2post_sum(self.g, self.post.num, self.post_ids, self.pre_ids)
      else:
        if self.weight_type == 'homo':
          post_g = (self.g_max * self.g) @ self.conn_mat
        else:
          post_g = self.g @ self.g_max

    # output
    self.post.input -= post_g * (self.post.V - self.E)


class GABAa(AMPA):
  r"""GABAa conductance-based synapse model.

  **Model Descriptions**

  GABAa synapse model has the same equation with the `AMPA synapse <./brainmodels.synapses.AMPA.rst>`_,

  .. math::

      \frac{d g}{d t}&=\alpha[T](1-g) - \beta g \\
      I_{syn}&= - g_{max} g (V - E)

  but with the difference of:

  - Reversal potential of synapse :math:`E` is usually low, typically -80. mV
  - Activating rate constant :math:`\alpha=0.53`
  - De-activating rate constant :math:`\beta=0.18`
  - Transmitter concentration :math:`[T]=1\,\mu ho(\mu S)` when synapse is
    triggered by a pre-synaptic spike, with the duration of 1. ms.

  **Model Examples**

  - `Gamma oscillation network model <https://brainpy-examples.readthedocs.io/en/latest/oscillation_synchronization/Wang_1996_gamma_oscillation.html>`_


  Parameters
  ----------
  pre: NeuGroup
    The pre-synaptic neuron group.
  post: NeuGroup
    The post-synaptic neuron group.
  conn: optional, ndarray, JaxArray, dict of (str, ndarray), TwoEndConnector
    The synaptic connections.
  conn_type: str
    The connection type used for model speed optimization. It can be
    `sparse` and `dense`. The default is `dense`.
  delay_step: int, ndarray, JaxArray, Initializer, Callable
    The delay length. It should be the value of :math:`\mathrm{delay\_time / dt}`.
  E: float
    The reversal potential for the synaptic current. [mV]
  g_max: float, ndarray, JaxArray, Initializer, Callable
    The synaptic strength (the maximum conductance). Default is 1.
  alpha: float
    Binding constant. Default 0.062
  beta: float
    Unbinding constant. Default 3.57
  T: float
    Transmitter concentration when synapse is triggered by
    a pre-synaptic spike.. Default 1 [mM].
  T_duration: float
    Transmitter concentration duration time after being triggered. Default 1 [ms]
  name: str
    The name of this synaptic projection.
  method: str
    The numerical integration methods.

  References
  ----------

  .. [1] Destexhe, Alain, and Denis Paré. "Impact of network activity
         on the integrative properties of neocortical pyramidal neurons
         in vivo." Journal of neurophysiology 81.4 (1999): 1531-1547.
  """

  def __init__(
      self,
      pre: NeuGroup,
      post: NeuGroup,
      conn: Union[TwoEndConnector, Tensor, Dict[str, Tensor]],
      conn_type: str = 'dense',
      g_max: Union[float, Tensor, Initializer, Callable] = 0.04,
      E: float = -80.,
      alpha: float = 0.53,
      beta: float = 0.18,
      T: float = 1.,
      T_duration: float = 1.,
      delay_step: Union[int, Tensor, Initializer, Callable] = None,
      method: str = 'exp_auto',
      name: str = None
  ):
    super(GABAa, self).__init__(pre, post, conn,
                                conn_type=conn_type,
                                delay_step=delay_step,
                                g_max=g_max,
                                E=E,
                                alpha=alpha,
                                beta=beta,
                                T=T,
                                T_duration=T_duration,
                                method=method,
                                name=name)



# -*- coding: utf-8 -*-

import warnings

from brainpy import optimizers

__all__ = [
  'SGD',
  'Momentum',
  'MomentumNesterov',
  'Adagrad',
  'Adadelta',
  'RMSProp',
  'Adam',

  'Constant',
  'ExponentialDecay',
  'InverseTimeDecay',
  'PolynomialDecay',
  'PiecewiseConstant',
]


def SGD(*args, **kwargs):
  """SGD optimizer.

  .. deprecated:: 2.1.0
     Please use "brainpy.optim.SGD" instead.
  """
  warnings.warn('Please use "brainpy.optim.SGD" instead. '
                '"brainpy.math.optimizers.SGD" is '
                'deprecated since version 2.0.3. ',
                DeprecationWarning)
  return optimizers.SGD(*args, **kwargs)


def Momentum(*args, **kwargs):
  """Momentum optimizer.

  .. deprecated:: 2.1.0
     Please use "brainpy.optim.Momentum" instead.
  """
  warnings.warn('Please use "brainpy.optim.Momentum" instead. '
                '"brainpy.math.optimizers.Momentum" is '
                'deprecated since version 2.0.3. ',
                DeprecationWarning)
  return optimizers.Momentum(*args, **kwargs)


def MomentumNesterov(*args, **kwargs):
  """MomentumNesterov optimizer.

  .. deprecated:: 2.1.0
     Please use "brainpy.optim.MomentumNesterov" instead.
  """
  warnings.warn('Please use "brainpy.optim.MomentumNesterov" instead. '
                '"brainpy.math.optimizers.MomentumNesterov" is '
                'deprecated since version 2.0.3. ',
                DeprecationWarning)
  return optimizers.MomentumNesterov(*args, **kwargs)


def Adagrad(*args, **kwargs):
  """Adagrad optimizer.

  .. deprecated:: 2.1.0
     Please use "brainpy.optim.Adagrad" instead.
  """
  warnings.warn('Please use "brainpy.optim.Adagrad" instead. '
                '"brainpy.math.optimizers.Adagrad" is '
                'deprecated since version 2.0.3. ',
                DeprecationWarning)
  return optimizers.Adagrad(*args, **kwargs)


def Adadelta(*args, **kwargs):
  """Adadelta optimizer.

  .. deprecated:: 2.1.0
     Please use "brainpy.optim.Adadelta" instead.
  """
  warnings.warn('Please use "brainpy.optim.Adadelta" instead. '
                '"brainpy.math.optimizers.Adadelta" is '
                'deprecated since version 2.0.3. ',
                DeprecationWarning)
  return optimizers.Adadelta(*args, **kwargs)


def RMSProp(*args, **kwargs):
  """RMSProp optimizer.

  .. deprecated:: 2.1.0
     Please use "brainpy.optim.RMSProp" instead.
  """
  warnings.warn('Please use "brainpy.optim.RMSProp" instead. '
                '"brainpy.math.optimizers.RMSProp" is '
                'deprecated since version 2.0.3. ',
                DeprecationWarning)
  return optimizers.RMSProp(*args, **kwargs)


def Adam(*args, **kwargs):
  """Adam optimizer.

  .. deprecated:: 2.1.0
     Please use "brainpy.optim.Adam" instead.
  """
  warnings.warn('Please use "brainpy.optim.Adam" instead. '
                '"brainpy.math.optimizers.Adam" is '
                'deprecated since version 2.0.3. ',
                DeprecationWarning)
  return optimizers.Adam(*args, **kwargs)


def Constant(*args, **kwargs):
  """Constant scheduler.

  .. deprecated:: 2.1.0
     Please use "brainpy.optim.Constant" instead.
  """
  warnings.warn('Please use "brainpy.optim.Constant" instead. '
                '"brainpy.math.optimizers.Constant" is '
                'deprecated since version 2.0.3. ',
                DeprecationWarning)
  return optimizers.Constant(*args, **kwargs)


def ExponentialDecay(*args, **kwargs):
  """ExponentialDecay scheduler.

  .. deprecated:: 2.1.0
     Please use "brainpy.optim.ExponentialDecay" instead.
  """
  warnings.warn('Please use "brainpy.optim.ExponentialDecay" instead. '
                '"brainpy.math.optimizers.ExponentialDecay" is '
                'deprecated since version 2.0.3. ',
                DeprecationWarning)
  return optimizers.ExponentialDecay(*args, **kwargs)


def InverseTimeDecay(*args, **kwargs):
  """InverseTimeDecay scheduler.

  .. deprecated:: 2.1.0
     Please use "brainpy.optim.InverseTimeDecay" instead.
  """
  warnings.warn('Please use "brainpy.optim.InverseTimeDecay" instead. '
                '"brainpy.math.optimizers.InverseTimeDecay" is '
                'deprecated since version 2.0.3. ',
                DeprecationWarning)
  return optimizers.InverseTimeDecay(*args, **kwargs)


def PolynomialDecay(*args, **kwargs):
  """PolynomialDecay scheduler.

  .. deprecated:: 2.1.0
     Please use "brainpy.optim.PolynomialDecay" instead.
  """
  warnings.warn('Please use "brainpy.optim.PolynomialDecay" instead. '
                '"brainpy.math.optimizers.PolynomialDecay" is '
                'deprecated since version 2.0.3. ',
                DeprecationWarning)
  return optimizers.PolynomialDecay(*args, **kwargs)


def PiecewiseConstant(*args, **kwargs):
  """PiecewiseConstant scheduler.

  .. deprecated:: 2.1.0
     Please use "brainpy.optim.PiecewiseConstant" instead.
  """
  warnings.warn('Please use "brainpy.optim.PiecewiseConstant" instead. '
                '"brainpy.math.optimizers.PiecewiseConstant" is '
                'deprecated since version 2.0.3. ',
                DeprecationWarning)
  return optimizers.PiecewiseConstant(*args, **kwargs)



# -*- coding: utf-8 -*-

"""
This module implements several loss functions.
"""

# - https://github.com/deepmind/optax/blob/master/optax/_src/loss.py
# - https://github.com/google/jaxopt/blob/main/jaxopt/_src/loss.py

import jax.numpy as jn
import jax.scipy
from jax.tree_util import tree_flatten

import brainpy.math as bm
from brainpy import errors

__all__ = [
  'cross_entropy_loss',
  'l1_loos',
  'l2_loss',
  'l2_norm',
  'huber_loss',
  'mean_absolute_error',
  'mean_squared_error',
  'mean_squared_log_error',
]

_reduction_error = 'Only support reduction of "mean", "sum" and "none", but we got "%s".'


def _return(outputs, reduction):
  if reduction == 'mean':
    return outputs.mean()
  elif reduction == 'sum':
    return outputs.sum()
  elif reduction == 'none':
    return outputs
  else:
    raise errors.UnsupportedError(_reduction_error % reduction)


def cross_entropy_loss(logits, targets, weight=None, reduction='mean'):
  r"""This criterion combines ``LogSoftmax`` and `NLLLoss`` in one single class.

  It is useful when training a classification problem with `C` classes.
  If provided, the optional argument :attr:`weight` should be a 1D `Tensor`
  assigning weight to each of the classes. This is particularly useful when
  you have an unbalanced training set.

  The ``input`` is expected to contain raw, unnormalized scores for each class.
  ``input`` has to be an array of size either :math:`(minibatch, C)` or
  :math:`(d_1, d_2, ..., d_K, minibatch, C)` with :math:`K \geq 1` for the
  `K`-dimensional case (described later).

  This criterion expects a class index in the range :math:`[0, C-1]` as the
  `target` for each value of a 1D tensor of size `minibatch`.

  The loss can be described as:

  .. math::
      \text{loss}(x, class) = -\log\left(\frac{\exp(x[class])}{\sum_j \exp(x[j])}\right)
                     = -x[class] + \log\left(\sum_j \exp(x[j])\right)

  or in the case of the :attr:`weight` argument being specified:

  .. math::
      \text{loss}(x, class) = weight[class] \left(-x[class] + \log\left(\sum_j \exp(x[j])\right)\right)

  Can also be used for higher dimension inputs, such as 2D images, by providing
  an input of size :math:`(d_1, d_2, ..., d_K, minibatch, C)` with :math:`K \geq 1`,
  where :math:`K` is the number of dimensions, and a target of appropriate shape.

  Parameters
  ----------
  logits : jmath.JaxArray
    :math:`(N, C)` where `C = number of classes`, or
    :math:`(d_1, d_2, ..., d_K, N, C)` with :math:`K \geq 1`
    in the case of `K`-dimensional loss.
  targets : jmath.JaxArray
    :math:`(N, C)` or :math:`(N)`  where each value is
    :math:`0 \leq \text{targets}[i] \leq C-1`, or
    :math:`(d_1, d_2, ..., d_K, N, C)` or :math:`(d_1, d_2, ..., d_K, N)`
    with :math:`K \geq 1` in the case of K-dimensional loss.
  weight : mjax.JaxArray, optional
    A manual rescaling weight given to each class. If given, has to be an array of size `C`.
  reduction : str, optional
    Specifies the reduction to apply to the output: ``'none'`` | ``'mean'`` | ``'sum'``.
    - ``'none'``: no reduction will be applied,
    - ``'mean'``: the weighted mean of the output is taken,
    - ``'sum'``: the output will be summed.

  Returns
  -------
  output : scalar, mjax.JaxArray
    If :attr:`reduction` is ``'none'``, then the same size as the target:
    :math:`(N)`, or  :math:`(d_1, d_2, ..., d_K, N)` with :math:`K \geq 1`
    in the case of K-dimensional loss.
  """
  targets = bm.as_device_array(targets)
  logits = bm.as_device_array(logits)

  # loss
  if bm.ndim(targets) + 1 == bm.ndim(logits):
    # targets_old = targets.reshape((-1,))
    # length = targets_old.shape[0]
    # rows = jn.arange(length)
    # targets = ops.zeros((length, logits.shape[-1]))
    # targets[rows, targets_old] = 1.
    # targets = targets.reshape(logits.shape).value
    targets = bm.activations.one_hot(targets, logits.shape[-1])
  loss = jax.scipy.special.logsumexp(logits, axis=-1) - (logits * targets).sum(axis=-1)

  # weighted loss
  if weight:
    loss *= weight[targets]
    raise NotImplementedError

  return _return(outputs=loss, reduction=reduction)


def cross_entropy_sparse(logits, labels):
  r"""Computes the softmax cross-entropy loss.

  Args:
      logits: (batch, ..., #class) tensor of logits.
      labels: (batch, ...) integer tensor of label indexes in {0, ...,#nclass-1} or just a single integer.

  Returns:
      (batch, ...) tensor of the cross-entropy for each entry.
  """

  if isinstance(labels, int):
    labeled_logits = logits[..., labels]
  else:
    logits = bm.as_device_array(logits)
    labels = bm.as_device_array(labels)
    labeled_logits = jn.take_along_axis(logits, labels, -1).squeeze(-1)
  loss = jax.scipy.special.logsumexp(logits, axis=-1) - labeled_logits
  return loss


def cross_entropy_sigmoid(logits, labels):
  """Computes the sigmoid cross-entropy loss.

  Args:
      logits: (batch, ..., #class) tensor of logits.
      labels: (batch, ..., #class) tensor of label probabilities (e.g. labels.sum(axis=-1) must be 1)

  Returns:
      (batch, ...) tensor of the cross-entropies for each entry.
  """
  return jax.numpy.maximum(logits, 0) - logits * labels + \
         jax.numpy.log(1 + jax.numpy.exp(-jax.numpy.abs(logits)))


def l1_loos(logits, targets, reduction='sum'):
  r"""Creates a criterion that measures the mean absolute error (MAE) between each element in
  the logits :math:`x` and targets :math:`y`. It is useful in regression problems.

  The unreduced (i.e. with :attr:`reduction` set to ``'none'``) loss can be described as:

  .. math::
      \ell(x, y) = L = \{l_1,\dots,l_N\}^\top, \quad
      l_n = \left| x_n - y_n \right|,

  where :math:`N` is the batch size. If :attr:`reduction` is not ``'none'``
  (default ``'mean'``), then:

  .. math::
      \ell(x, y) =
      \begin{cases}
          \operatorname{mean}(L), & \text{if reduction} = \text{`mean';}\\
          \operatorname{sum}(L),  & \text{if reduction} = \text{`sum'.}
      \end{cases}

  :math:`x` and :math:`y` are tensors of arbitrary shapes with a total
  of :math:`n` elements each.

  The sum operation still operates over all the elements, and divides by :math:`n`.

  The division by :math:`n` can be avoided if one sets ``reduction = 'sum'``.

  Supports real-valued and complex-valued inputs.

  Parameters
  ----------
  logits : jmath.JaxArray
    :math:`(N, *)` where :math:`*` means, any number of additional dimensions.
  targets : jmath.JaxArray
    :math:`(N, *)`, same shape as the input.
  reduction : str
    Specifies the reduction to apply to the output: ``'none'`` | ``'mean'`` | ``'sum'``.
    Default: ``'mean'``.
    - ``'none'``: no reduction will be applied,
    - ``'mean'``: the sum of the output will be divided by the number of elements in the output,
    - ``'sum'``: the output will be summed. Note: :attr:`size_average`

  Returns
  -------
  output : scalar.
    If :attr:`reduction` is ``'none'``, then :math:`(N, *)`, same shape as the input.
  """
  diff = (logits - targets).reshape((logits.shape[0], -1))
  norm = jn.linalg.norm(bm.as_device_array(diff), ord=1, axis=1, keepdims=False)
  return _return(outputs=norm, reduction=reduction)


def l2_loss(predicts, targets):
  r"""Computes the L2 loss.

  The 0.5 term is standard in "Pattern Recognition and Machine Learning"
  by Bishop [1]_, but not "The Elements of Statistical Learning" by Tibshirani.

  Parameters
  ----------

  predicts: JaxArray
    A vector of arbitrary shape.
  targets: JaxArray
    A vector of shape compatible with predictions.

  Returns
  -------
  loss : float
    A scalar value containing the l2 loss.

  References
  ----------
  .. [1] Bishop, Christopher M. 2006. Pattern Recognition and Machine Learning.
  """
  return bm.as_device_array(0.5 * (predicts - targets) ** 2)


def l2_norm(x):
  """Computes the L2 loss.

  Args:
      x: n-dimensional tensor of floats.

  Returns:
      scalar tensor containing the l2 loss of x.
  """
  leaves, _ = tree_flatten(x)
  return jn.sqrt(sum(jn.vdot(x, x) for x in leaves))


def mean_absolute_error(x, y, axis=None):
  r"""Computes the mean absolute error between x and y.

  Args:
      x: a tensor of shape (d0, .. dN-1).
      y: a tensor of shape (d0, .. dN-1).
      keep_axis: a sequence of the dimensions to keep, use `None` to return a scalar value.

  Returns:
      tensor of shape (d_i, ..., for i in keep_axis) containing the mean absolute error.
  """
  r = bm.abs(x - y)
  return jn.mean(bm.as_device_array(r), axis=axis)


def mean_squared_error(predicts, targets, axis=None):
  r"""Computes the mean squared error between x and y.

  Args:
      predicts: a tensor of shape (d0, .. dN-1).
      targets: a tensor of shape (d0, .. dN-1).
      keep_axis: a sequence of the dimensions to keep, use `None` to return a scalar value.

  Returns:
      tensor of shape (d_i, ..., for i in keep_axis) containing the mean squared error.
  """
  r = (predicts - targets) ** 2
  return jn.mean(bm.as_device_array(r), axis=axis)


def mean_squared_log_error(y_true, y_pred, axis=None):
  r"""Computes the mean squared logarithmic error between y_true and y_pred.

  Args:
      y_true: a tensor of shape (d0, .. dN-1).
      y_pred: a tensor of shape (d0, .. dN-1).
      keep_axis: a sequence of the dimensions to keep, use `None` to return a scalar value.

  Returns:
      tensor of shape (d_i, ..., for i in keep_axis) containing the mean squared error.
  """
  r = (bm.log1p(y_true) - bm.log1p(y_pred)) ** 2
  return jn.mean(bm.as_device_array(r), axis=axis)


def huber_loss(predicts, targets, delta: float = 1.0):
  r"""Huber loss.

  Huber loss is similar to L2 loss close to zero, L1 loss away from zero.
  If gradient descent is applied to the `huber loss`, it is equivalent to
  clipping gradients of an `l2_loss` to `[-delta, delta]` in the backward pass.

  Parameters
  ----------
  predicts: JaxArray
    predictions
  targets: JaxArray
    ground truth
  delta: float
    radius of quadratic behavior

  Returns
  -------
  loss : float
    The loss value.

  References
  ----------
  .. [1] https://en.wikipedia.org/wiki/Huber_loss
  """
  diff = bm.as_device_array(bm.abs(targets - predicts))
  # 0.5 * err^2                  if |err| <= d
  # 0.5 * d^2 + d * (|err| - d)  if |err| > d
  return jn.where(diff > delta, delta * (diff - .5 * delta), 0.5 * diff ** 2)


def binary_logistic_loss(logits: float, labels: int, ) -> float:
  """Binary logistic loss.

  Args:
    labels: ground-truth integer label (0 or 1).
    logits: score produced by the model (float).
  Returns:
    loss value
  """
  # Softplus is the Fenchel conjugate of the Fermi-Dirac negentropy on [0, 1].
  # softplus = proba * logit - xlogx(proba) - xlogx(1 - proba),
  # where xlogx(proba) = proba * log(proba).
  return bm.activations.softplus(logits) - labels * logits


def multiclass_logistic_loss(label: int, logits: jn.ndarray) -> float:
  """Multiclass logistic loss.

  Args:
    label: ground-truth integer label, between 0 and n_classes - 1.
    logits: scores produced by the model, shape = (n_classes, ).
  Returns:
    loss value
  """
  n_classes = logits.shape[0]
  one_hot = jax.nn.one_hot(label, n_classes)
  # Logsumexp is the Fenchel conjugate of the Shannon negentropy on the simplex.
  # logsumexp = jnp.dot(proba, logits) - jnp.dot(proba, jnp.log(proba))
  return jax.scipy.special.logsumexp(logits) - jn.dot(logits, one_hot)


def smooth_labels(labels, alpha: float) -> jn.ndarray:
  r"""Apply label smoothing.
  Label smoothing is often used in combination with a cross-entropy loss.
  Smoothed labels favour small logit gaps, and it has been shown that this can
  provide better model calibration by preventing overconfident predictions.
  References:
    [Müller et al, 2019](https://arxiv.org/pdf/1906.02629.pdf)
  Args:
    labels: one hot labels to be smoothed.
    alpha: the smoothing factor, the greedy category with be assigned
      probability `(1-alpha) + alpha / num_categories`
  Returns:
    a smoothed version of the one hot input labels.
  """
  num_categories = labels.shape[-1]
  return (1.0 - alpha) * labels + alpha / num_categories


def sigmoid_binary_cross_entropy(logits, labels):
  """Computes sigmoid cross entropy given logits and multiple class labels.
  Measures the probability error in discrete classification tasks in which
  each class is an independent binary prediction and different classes are
  not mutually exclusive. This may be used for multilabel image classification
  for instance a model may predict that an image contains both a cat and a dog.
  References:
    [Goodfellow et al, 2016](http://www.deeplearningbook.org/contents/prob.html)
  Args:
    logits: unnormalized log probabilities.
    labels: the probability for that class.
  Returns:
    a sigmoid cross entropy loss.
  """
  log_p = jax.nn.log_sigmoid(logits)
  # log(1 - sigmoid(x)) = log_sigmoid(-x), the latter more numerically stable
  log_not_p = jax.nn.log_sigmoid(-logits)
  return -labels * log_p - (1. - labels) * log_not_p


def softmax_cross_entropy(logits, labels):
  """Computes the softmax cross entropy between sets of logits and labels.
  Measures the probability error in discrete classification tasks in which
  the classes are mutually exclusive (each entry is in exactly one class).
  For example, each CIFAR-10 image is labeled with one and only one label:
  an image can be a dog or a truck, but not both.
  References:
    [Goodfellow et al, 2016](http://www.deeplearningbook.org/contents/prob.html)
  Args:
    logits: unnormalized log probabilities.
    labels: a valid probability distribution (non-negative, sum to 1), e.g a
      one hot encoding of which class is the correct one for each input.
  Returns:
    the cross entropy loss.
  """
  logits = bm.as_device_array(logits)
  labels = bm.as_device_array(labels)
  return -jn.sum(labels * jax.nn.log_softmax(logits, axis=-1), axis=-1)


def log_cosh(predicts, targets=None, ):
  r"""Calculates the log-cosh loss for a set of predictions.

  log(cosh(x)) is approximately `(x**2) / 2` for small x and `abs(x) - log(2)`
  for large x.  It is a twice differentiable alternative to the Huber loss.
  References:
    [Chen et al, 2019](https://openreview.net/pdf?id=rkglvsC9Ym)
  Args:
    predicts: a vector of arbitrary shape.
    targets: a vector of shape compatible with predictions; if not provided
      then it is assumed to be zero.
  Returns:
    the log-cosh loss.
  """
  errors = (predicts - targets) if (targets is not None) else predicts
  errors = bm.as_device_array(errors)
  # log(cosh(x)) = log((exp(x) + exp(-x))/2) = log(exp(x) + exp(-x)) - log(2)
  return jn.logaddexp(errors, -errors) - jn.log(2.0).astype(errors.dtype)



# -*- coding: utf-8 -*-


"""
This module provide basic Node class for whole ``brainpy.nn`` system.

- ``brainpy.nn.Node``: The fundamental class representing the node or the element.
- ``brainpy.nn.RecurrentNode``: The recurrent node which has a self-connection.
- ``brainpy.nn.Network``: The network model which is composed of multiple node elements.
  Once the Network instance receives a node operation, the wrapped elements, the new
  elements, and their connection edges will be formed as another Network instance.
  This means ``brainpy.nn.Network`` is only used to pack element nodes. It will be
  never be an element node.
- ``brainpy.nn.FrozenNetwork``: The whole network which can be represented as a basic
  elementary node when composing a larger network (TODO).
"""

from copy import copy, deepcopy
from typing import (Dict, Sequence, Tuple, Union, Optional, Any, Callable)

import jax.numpy as jnp

from brainpy import tools, math as bm
from brainpy.base import Base, Collector
from brainpy.errors import (UnsupportedError,
                            PackageMissingError,
                            ModelBuildError,
                            MathError)
from brainpy.nn.algorithms.offline import OfflineAlgorithm
from brainpy.nn.algorithms.online import OnlineAlgorithm
from brainpy.nn.datatypes import (DataType, SingleData, MultipleData)
from brainpy.nn.graph_flow import (find_senders_and_receivers,
                                   find_entries_and_exits,
                                   detect_cycle,
                                   detect_path)
from brainpy.tools.checking import (check_dict_data,
                                    check_shape_except_batch,
                                    check_integer)
from brainpy.types import Tensor

operations = None

__all__ = [
  'Node', 'Network',
  'RecurrentNode',  # a marker for recurrent node
  'FrozenNetwork',  # a marker for frozen network
]

NODE_STATES = ['inputs', 'feedbacks', 'state', 'output']

SUPPORTED_LAYOUTS = ['shell_layout',
                     'multipartite_layout',
                     'spring_layout',
                     'spiral_layout',
                     'spectral_layout',
                     'random_layout',
                     'planar_layout',
                     'kamada_kawai_layout',
                     'circular_layout']


def not_implemented(fun: Callable) -> Callable:
  """Marks the given module method is not implemented.

  Methods wrapped in @not_implemented can define submodules directly within the method.

  For instance::

    @not_implemented
    init_fb(self):
      ...

    @not_implemented
    def feedback(self):
      ...
  """
  fun.not_implemented = True
  return fun


class Node(Base):
  """Basic Node class for neural network building in BrainPy."""

  '''Support multiple types of data pass, including "PassOnlyOne" (by default), 
  "PassSequence", "PassNameDict", etc. and user-customized type which inherits 
  from basic "SingleData" or "MultipleData". 
  
  This setting will change the feedforward/feedback input data which pass into 
  the "call()" function and the sizes of the feedforward/feedback input data.'''
  data_pass = SingleData()

  '''Offline fitting method.'''
  offline_fit_by: Union[Callable, OfflineAlgorithm]

  '''Online fitting method.'''
  online_fit_by: OnlineAlgorithm

  def __init__(
      self,
      name: Optional[str] = None,
      input_shape: Optional[Union[Sequence[int], int]] = None,
      trainable: bool = True
  ):

    # initialize parameters
    self._feedforward_shapes = None  # input shapes
    self._output_shape = None  # output size
    self._feedback_shapes = None  # feedback shapes
    self._is_ff_initialized = False
    self._is_fb_initialized = False
    self._is_state_initialized = False
    self._is_fb_state_initialized = False
    self._trainable = trainable
    self._state = None  # the state of the current node
    self._fb_output = None  # the feedback output of the current node
    # data pass
    if not isinstance(self.data_pass, DataType):
      raise ValueError(f'Unsupported data pass type {type(self.data_pass)}. '
                       f'Only support {DataType.__class__}')

    # super initialization
    super(Node, self).__init__(name=name)

    # parameters
    if input_shape is not None:
      self._feedforward_shapes = {self.name: (None,) + tools.to_size(input_shape)}

  def __repr__(self):
    return (f"{type(self).__name__}(name={self.name}, "
            f"forwards={self.feedforward_shapes}, "
            f"feedbacks={self.feedback_shapes}, "
            f"output={self.output_shape})")

  def __call__(self, *args, **kwargs) -> Tensor:
    """The main computation function of a Node.

    Parameters
    ----------
    ff: dict, sequence, JaxArray, ndarray
      The feedforward inputs.
    fb: optional, dict, sequence, JaxArray, ndarray
      The feedback inputs.
    forced_states: optional, dict
      The fixed state for the nodes in the network.
    forced_feedbacks: optional, dict
      The fixed feedback for the nodes in the network.
    monitors: optional, sequence
      Can be used to monitor the state or the attribute of a node in the network.
    **kwargs
      Other parameters which will be parsed into every node.

    Returns
    -------
    Tensor
      A output tensor value, or a dict of output tensors.
    """
    return self._call(*args, **kwargs)

  def __rshift__(self, other):  # "self >> other"
    global operations
    if operations is None: from . import operations
    return operations.ff_connect(self, other)

  def __rrshift__(self, other):  # "other >> self"
    global operations
    if operations is None: from . import operations
    return operations.ff_connect(other, self)

  def __irshift__(self, other):  # "self >>= other"
    raise ValueError('Only Network objects support inplace feedforward connection.')

  def __lshift__(self, other):  # "self << other"
    global operations
    if operations is None: from . import operations
    return operations.fb_connect(other, self)

  def __rlshift__(self, other):  # "other << self"
    global operations
    if operations is None: from . import operations
    return operations.fb_connect(self, other)

  def __ilshift__(self, other):  # "self <<= other"
    raise ValueError('Only Network objects support inplace feedback connection.')

  def __and__(self, other):  # "self & other"
    global operations
    if operations is None: from . import operations
    return operations.merge(self, other)

  def __rand__(self, other):  # "other & self"
    global operations
    if operations is None: from . import operations
    return operations.merge(other, self)

  def __iand__(self, other):
    raise ValueError('Only Network objects support inplace merging.')

  def __getitem__(self, item):  # like "[:10]"
    if isinstance(item, str):
      raise ValueError('Node only supports slice, not retrieve by the name.')
    else:
      global operations
      if operations is None: from . import operations
      return operations.select(self, item)

  @property
  def state(self) -> Optional[Tensor]:
    """Node current internal state."""
    if self._is_ff_initialized:
      return self._state
    return None

  @state.setter
  def state(self, value: Tensor):
    raise NotImplementedError('Please use "set_state()" to reset the node state, '
                              'or use "self.state.value" to change the state content.')

  def set_state(self, state):
    """
    Safely set the state of the node.

    This method allows the maximum flexibility to change the
    node state. It can set a new data (same shape, same dtype)
    to the state. It can also set a new data with the different
    shape. We highly recommend the user to use this function.
    instead of using ``self.state.value``.
    """
    if self.state is None:
      if self.output_shape is not None:
        check_shape_except_batch(self.output_shape, state.shape)
      self._state = bm.Variable(state) if not isinstance(state, bm.Variable) else state
    else:
      check_shape_except_batch(self.state.shape, state.shape)
      if self.state.dtype != state.dtype:
        raise MathError('Cannot set the state, because the dtype is not consistent: '
                        f'{self.state.dtype} != {state.dtype}')
      self.state._value = bm.as_device_array(state)

  @property
  def fb_output(self) -> Optional[Tensor]:
    return self._fb_output

  @fb_output.setter
  def fb_output(self, value: Tensor):
    raise NotImplementedError('Please use "set_fb_output()" to reset the node feedback state, '
                              'or use "self.fb_output.value" to change the state content.')

  def set_fb_output(self, state: Tensor):
    """
    Safely set the feedback state of the node.

    This method allows the maximum flexibility to change the
    node state. It can set a new data (same shape, same dtype)
    to the state. It can also set a new data with the different
    shape. We highly recommend the user to use this function.
    instead of using ``self.fb_output.value``.
    """
    if self.fb_output is None:
      if self.output_shape is not None:
        check_shape_except_batch(self.output_shape, state.shape)
      self._fb_output = bm.Variable(state) if not isinstance(state, bm.Variable) else state
    else:
      check_shape_except_batch(self.fb_output.shape, state.shape)
      if self.fb_output.dtype != state.dtype:
        raise MathError('Cannot set the feedback state, because the dtype is '
                        f'not consistent: {self.fb_output.dtype} != {state.dtype}')
      self.fb_output._value = bm.as_device_array(state)

  @property
  def trainable(self) -> bool:
    """Returns if the Node can be trained."""
    return self._trainable

  @property
  def is_initialized(self) -> bool:
    if self._is_ff_initialized and self._is_state_initialized:
      if self.feedback_shapes is not None:
        if self._is_fb_initialized and self._is_fb_state_initialized:
          return True
        else:
          return False
      else:
        return True
    else:
      return False

  @trainable.setter
  def trainable(self, value: bool):
    """Freeze or unfreeze the Node. If set to False,
    learning is stopped."""
    assert isinstance(value, bool), 'Must be a boolean.'
    self._trainable = value

  @property
  def feedforward_shapes(self):
    """Input data size."""
    return self.data_pass.filter(self._feedforward_shapes)

  @feedforward_shapes.setter
  def feedforward_shapes(self, size):
    self.set_feedforward_shapes(size)

  def set_feedforward_shapes(self, feedforward_shapes: Dict):
    if not self._is_ff_initialized:
      check_dict_data(feedforward_shapes,
                      key_type=(Node, str),
                      val_type=(list, tuple),
                      name='feedforward_shapes')
      self._feedforward_shapes = feedforward_shapes
    else:
      if self.feedforward_shapes is not None:
        sizes1 = sorted(list(self._feedforward_shapes.values()))
        sizes2 = sorted(list(feedforward_shapes.values()))
        if sizes1 != sizes2:
          raise ValueError(f"Impossible to reset the input shapes of {self.name}. "
                           f"Because this Node has the input shapes {sizes1}. "
                           f"While we got input shapes of {sizes2}")
        self._feedforward_shapes = feedforward_shapes

  @property
  def feedback_shapes(self):
    """Output data size."""
    return self.data_pass.filter(self._feedback_shapes)

  @feedback_shapes.setter
  def feedback_shapes(self, size):
    self.set_feedback_shapes(size)

  def set_feedback_shapes(self, fb_shapes: Dict):
    if not self._is_fb_initialized:
      check_dict_data(fb_shapes,
                      key_type=(Node, str),
                      val_type=(tuple, list),
                      name='fb_shapes')
      self._feedback_shapes = fb_shapes
    else:
      if self.feedback_shapes is not None:
        sizes1 = sorted(list(self._feedback_shapes.values()))
        sizes2 = sorted(list(fb_shapes.values()))
        if sizes1 != sizes2:
          raise ValueError(f"Impossible to reset the feedback shapes of {self.name}. "
                           f"Because this Node has the feedback shapes {sizes1}. "
                           f"While we got feedback shapes of {sizes2}")
        self._feedback_shapes = fb_shapes

  @property
  def output_shape(self) -> Optional[Tuple[int]]:
    """Output data size."""
    return self._output_shape

  @output_shape.setter
  def output_shape(self, size):
    self.set_output_shape(size)

  @property
  def is_feedback_input_supported(self):
    if hasattr(self.init_fb_conn, 'not_implemented'):
      if self.init_fb_conn.not_implemented:
        return False
    return True

  @property
  def is_feedback_supported(self):
    if self.fb_output is None:
      return False
    else:
      return True

  def set_output_shape(self, shape: Sequence[int]):
    if not self._is_ff_initialized:
      if not isinstance(shape, (tuple, list)):
        raise ValueError(f'Must be a sequence of int, but got {shape}')
      self._output_shape = tuple(shape)
    else:
      check_shape_except_batch(shape, self.output_shape)

  def nodes(self, method='absolute', level=1, include_self=True):
    return super(Node, self).nodes(method=method, level=level, include_self=include_self)

  def vars(self, method='absolute', level=1, include_self=True):
    return super(Node, self).vars(method=method, level=level, include_self=include_self)

  def train_vars(self, method='absolute', level=1, include_self=True):
    return super(Node, self).train_vars(method=method, level=level, include_self=include_self)

  def copy(self,
           name: str = None,
           shallow: bool = False):
    """Returns a copy of the Node.

    Parameters
    ----------
    name : str
        Name of the Node copy.
    shallow : bool, default to False
        If False, performs a deep copy of the Node.

    Returns
    -------
    Node
        A copy of the Node.
    """
    if shallow:
      new_obj = copy(self)
    else:
      new_obj = deepcopy(self)
    new_obj.name = self.unique_name(name or (self.name + '_copy'))
    return new_obj

  def _init_ff_conn(self):
    if not self._is_ff_initialized:
      self.init_ff_conn()
      if self.output_shape is None:
        raise ValueError(f'Please set the output shape when implementing '
                         f'"init_ff_conn()" of the node {self.name}')
      self._is_ff_initialized = True

  def _init_fb_conn(self):
    if not self._is_fb_initialized:
      try:
        self.init_fb_conn()
      except Exception as e:
        raise ModelBuildError(f"{self.name} initialization failed.") from e
      self._is_fb_initialized = True

  @not_implemented
  def init_fb_conn(self):
    """Initialize the feedback connections.
    This function will be called only once."""
    raise ValueError(f'This node \n\n{self} \n\ndoes not support feedback connection.')

  def init_ff_conn(self):
    """Initialize the feedforward connections.
    This function will be called only once."""
    raise NotImplementedError('Please implement the feedforward initialization.')

  def _init_state(self, num_batch=1):
    state = self.init_state(num_batch)
    if state is not None:
      self.set_state(state)
    self._is_state_initialized = True

  def _init_fb_output(self, num_batch=1):
    output = self.init_fb_output(num_batch)
    if output is not None:
      self.set_fb_output(output)
    self._is_fb_state_initialized = True

  def init_state(self, num_batch=1) -> Optional[Tensor]:
    """Set the initial node state.

    This function can be called multiple times."""
    pass

  def init_fb_output(self, num_batch=1) -> Optional[Tensor]:
    """Set the initial node feedback state.

    This function can be called multiple times. However,
    it is only triggered when the node has feedback connections.
    """
    return bm.zeros((num_batch,) + self.output_shape[1:], dtype=bm.float_)

  def initialize(self, num_batch: int = 1):
    """
    Initialize the node. This function must be called before applying JIT.

    This function is useful, because it is independent of the __call__ function.
    We can use this function before we apply JIT to __call__ function.
    """

    # feedforward initialization
    if self.feedforward_shapes is None:
      raise ValueError('Cannot initialize this node, because we detect '
                       'both "feedforward_shapes" is None. '
                       'Two ways can solve this problem:\n\n'
                       '1. Connecting an instance of "brainpy.nn.Input()" to this node. \n'
                       '2. Providing the "input_shape" when initialize the node.')
    check_integer(num_batch, 'num_batch', min_bound=0, allow_none=False)
    self._init_ff_conn()

    # initialize state
    self._init_state(num_batch)

    if self.feedback_shapes is not None:
      # feedback initialization
      self._init_fb_conn()
      # initialize feedback state
      self._init_fb_output(num_batch)

  def _check_inputs(self, ff, fb=None):
    # check feedforward inputs
    if isinstance(ff, (bm.ndarray, jnp.ndarray)):
      ff = {self.name: ff}
    if not isinstance(ff, dict):
      raise ValueError(f'"ff" must be a dict or a tensor, got {type(ff)}: {ff}')
    if self.name not in ff:
      raise ValueError(f'Cannot find input for this node {self} when given "ff" {ff}')
    for k, size in self._feedforward_shapes.items():
      if k not in ff:
        raise ValueError(f"The required key {k} is not provided in feedforward inputs.")
      check_shape_except_batch(size, ff[k].shape)
    if self.state is not None:
      for inp in ff.values():
        if self.state.shape[0] != inp.shape[0]:
          raise ValueError(f'The batch size of the input data {inp.shape[0]} is not '
                           f'equal to the batch size of the node state {self.state.shape[0]}. '
                           f'Maybe you need to reinitialize the data with the desired '
                           f'batch size by ".initialize(num_batch)", or change the data '
                           f'consistent with the data batch size {self.state.shape[0]}.')

    # check feedback inputs
    if fb is not None:
      if not isinstance(fb, dict):
        raise ValueError(f'"fb" must be a dict, got {type(fb)}: {fb}')
      # check feedback consistency
      for k, size in self._feedback_shapes.items():
        if k not in fb:
          raise ValueError(f"The required key {k} is not provided in feedback inputs.")
        check_shape_except_batch(size, fb[k].shape)
      if self.state is not None:
        for inp in fb.values():
          if self.state.shape[0] != inp.shape[0]:
            raise ValueError(f'The batch size of the feedback data {inp.shape[0]} is not '
                             f'equal to the batch size of the node state {self.state.shape[0]}. '
                             f'Maybe you need to reinitialize the data with the desired '
                             f'batch size by ".initialize(num_batch)", or change the data '
                             f'consistent with the data batch size {self.state.shape[0]}.')
    # data
    ff = self.data_pass.filter(ff)
    fb = self.data_pass.filter(fb)
    return ff, fb

  def _call(self,
            ff: Union[Tensor, Dict[Any, Tensor]],
            fb: Optional[Union[Tensor, Dict[Any, Tensor]]] = None,
            forced_states: Dict[str, Tensor] = None,
            forced_feedbacks: Dict[str, Tensor] = None,
            monitors=None,
            **kwargs) -> Union[Tensor, Tuple[Tensor, Dict]]:
    if not self.is_initialized:
      raise ValueError('Please initialize the Node first by calling "initialize()" function.')

    # initialize the forced data
    if forced_states is None:
      forced_states = dict()
    if isinstance(forced_states, (bm.ndarray, jnp.ndarray)):
      forced_states = {self.name: forced_states}
    check_dict_data(forced_states, key_type=str, val_type=(bm.ndarray, jnp.ndarray))
    if forced_feedbacks is not None:
      if len(forced_feedbacks) != 0:
        raise ValueError('Single instance of brainpy.nn.Node do '
                         'not support "forced_feedbacks"')
    # monitors
    need_return_monitor = True
    if monitors is None:
      monitors = tuple()
      need_return_monitor = False
    attr_monitors: Dict[str, Tensor] = {}
    state_monitors: Dict[str, Tensor] = {}
    for key in monitors:
      splits = key.split('.')
      if len(splits) != 2:
        raise ValueError(f'Every term in "monitors" must be (node.item), '
                         f'while we got {key}')
      if splits[0] not in self.implicit_nodes:
        raise ValueError(f'Cannot found the node {splits[0]}, this network '
                         f'only has {list(self.implicit_nodes.keys())}.')

      if splits[1] not in NODE_STATES:  # attribute monitor
        if not hasattr(self.implicit_nodes[splits[0]], splits[1]):
          raise UnsupportedError(f'Each node can monitor its states (including {NODE_STATES}), '
                                 f'or its attribute. While {splits[1]} is neither the state nor '
                                 f'the attribute of node {splits[0]}.')
        else:
          attr_monitors[key] = getattr(self.implicit_nodes[splits[0]], splits[1])
      else:  # state monitor
        if splits[1] == 'state':
          assert self.implicit_nodes[splits[0]].state is not None, (f'{splits[0]} has no state, while '
                                                                    f'the user try to monitor it.')
        state_monitors[key] = None

      if not isinstance(key, str):
        raise ValueError(f'"extra_returns" must be a sequence of string, '
                         f'while we got {type(key)}')
      splits = key.split('.')
      if len(splits) != 2:
        raise ValueError(f'Every term in "monitors" must be (node.item), '
                         f'while we got {key}')
      if splits[0] != self.name:
        raise ValueError(f"Cannot found the node {splits[0]}, this name of "
                         f"this node is {self.name}.")
      if splits[1] not in NODE_STATES:  # monitor attributes
        if not hasattr(self, key):
          raise UnsupportedError(f'Each node can monitor its states (including {NODE_STATES}), '
                                 f'or its attribute. While {key} is neither the state nor '
                                 f'the attribute of node \n\n{self}.')
        else:
          attr_monitors[key] = getattr(self, key)
      else:  # monitor states
        if splits[1] == 'state':
          if self.state is None:
            raise ValueError(f'{self} \n\nhas no state, while '
                             f'the user try to monitor its state.')
        state_monitors[key] = None

    # checking
    ff, fb = self._check_inputs(ff, fb=fb)

    # monitoring
    if f'{self.name}.inputs' in state_monitors:
      state_monitors[f'{self.name}.inputs'] = ff
    if f'{self.name}.feedbacks' in state_monitors:
      state_monitors[f'{self.name}.feedbacks'] = fb

    # forward pass
    output = self.forward(ff, fb, **kwargs)

    # monitoring
    if f'{self.name}.output' in state_monitors:
      state_monitors[f'{self.name}.output'] = output
    if f'{self.name}.state' in state_monitors:
      state_monitors[f'{self.name}.state'] = self.state
    attr_monitors.update(state_monitors)

    # outputs
    if need_return_monitor:
      return output, attr_monitors
    else:
      return output

  def forward(self, ff, fb=None, **shared_kwargs):
    """The feedforward computation function of a node.

    Parameters
    ----------
    ff: tensor, dict, sequence
      The feedforward inputs.
    fb: optional, tensor, dict, sequence
      The feedback inputs.
    **shared_kwargs
      Other parameters.

    Returns
    -------
    Tensor
      A output tensor value.
    """
    raise NotImplementedError

  def feedback(self, ff_output, **shared_kwargs):
    """The feedback computation function of a node.

    Parameters
    ----------
    ff_output: JaxArray
      The feedforward output when calling ``forward()`` function.
    **shared_kwargs
      Other global parameters.

    Returns
    -------
    Tensor
      A feedback output tensor value.
    """
    return ff_output

  @not_implemented
  def offline_fit(self, targets, ffs, fbs=None):
    """Offline training interface."""
    raise ValueError(f'This node \n\n{self} \n\ndoes not support offline training.')

  @not_implemented
  def online_init(self):
    """Online training initialization interface."""
    raise ValueError(f'This node \n\n{self} \n\ndoes not support online training.')

  @not_implemented
  def online_fit(self, target, ff, fb=None):
    """Online training fitting interface."""
    raise ValueError(f'This node \n\n{self} \n\ndoes not support online training.')


class RecurrentNode(Node):
  """
  Basic class for recurrent node.

  The supports for the recurrent node are:

  - Self-connection when using ``plot_node_graph()`` function
  - Set trainable state with ``state_trainable=True``.
  """

  def __init__(self,
               name: Optional[str] = None,
               input_shape: Optional[Union[Sequence[int], int]] = None,
               trainable: bool = False,
               state_trainable: bool = False):
    self._state_trainable = state_trainable
    self._train_state = None
    super(RecurrentNode, self).__init__(name=name,
                                        input_shape=input_shape,
                                        trainable=trainable)

  @property
  def state_trainable(self) -> bool:
    """Returns if the Node can be trained."""
    return self._state_trainable

  @property
  def train_state(self):
    return self._train_state

  def set_state(self, state):
    """Safely set the state of the node.

    This method allows the maximum flexibility to change the
    node state. It can set a new data (same shape, same dtype)
    to the state. It can also set the data with another batch size.

    We highly recommend the user to use this function.
    """
    if self.state is None:
      if self.output_shape is not None:
        check_shape_except_batch(self.output_shape, state.shape)
      self._state = bm.Variable(state) if not isinstance(state, bm.Variable) else state
      if self.state_trainable:
        self._train_state = bm.TrainVar(self._state[0])  # get the first elements as the initial state
        self._state[:] = self._train_state  # set all batch states the same
    else:
      check_shape_except_batch(self.state.shape, state.shape)
      if self.state.dtype != state.dtype:
        raise MathError('Cannot set the state, because the dtype is not consistent: '
                        f'{self.state.dtype} != {state.dtype}')
      if self.state_trainable:
        # get the batch size information
        state = bm.repeat(bm.expand_dims(self.train_state, axis=0), state.shape[0], axis=0)
        # set the state
        self.state._value = bm.as_device_array(state)
      else:
        self.state._value = bm.as_device_array(state)


class Network(Node):
  """Basic Network class for neural network building in BrainPy."""

  data_pass = MultipleData('sequence')

  def __init__(self,
               nodes: Optional[Sequence[Node]] = None,
               ff_edges: Optional[Sequence[Tuple[Node]]] = None,
               fb_edges: Optional[Sequence[Tuple[Node]]] = None,
               **kwargs):
    super(Network, self).__init__(**kwargs)
    # nodes (with tuple/list format)
    if nodes is None:
      self._nodes = tuple()
    else:
      self._nodes = tuple(nodes)
    # feedforward edges
    if ff_edges is None:
      self._ff_edges = tuple()
    else:
      self._ff_edges = tuple(ff_edges)
    # feedback edges
    if fb_edges is None:
      self._fb_edges = tuple()
    else:
      self._fb_edges = tuple(fb_edges)
    # initialize network
    self._network_init()

  def _network_init(self):
    # detect input and output nodes
    self._entry_nodes, self._exit_nodes = find_entries_and_exits(self._nodes, self._ff_edges)
    # build feedforward connection graph
    self._ff_senders, self._ff_receivers = find_senders_and_receivers(self._ff_edges)
    # build feedback connection graph
    self._fb_senders, self._fb_receivers = find_senders_and_receivers(self._fb_edges)
    # register nodes for brainpy.Base object
    self.implicit_nodes = Collector({n.name: n for n in self._nodes})
    # set initialization states
    self._is_initialized = False
    self._is_fb_initialized = False

  def __repr__(self):
    return f"{type(self).__name__}({', '.join([n.name for n in self._nodes])})"

  def __irshift__(self, other):  # "self >>= other"
    global operations
    if operations is None: from . import operations
    return operations.ff_connect(self, other, inplace=True)

  def __ilshift__(self, other):  # "self <<= other"
    global operations
    if operations is None: from . import operations
    return operations.fb_connect(self, other, inplace=True)

  def __iand__(self, other):
    global operations
    if operations is None: from . import operations
    return operations.merge(self, other, inplace=True)

  def __getitem__(self, item):
    if isinstance(item, str):
      return self.get_node(item)
    else:
      global operations
      if operations is None: from . import operations
      return operations.select(self, item)

  def get_node(self, name):
    if name in self.implicit_nodes:
      return self.implicit_nodes[name]
    else:
      raise KeyError(f"No node named '{name}' found in model {self.name}.")

  def nodes(self, method='absolute', level=1, include_self=False):
    return super(Node, self).nodes(method=method, level=level, include_self=include_self)

  @property
  def trainable(self) -> bool:
    """Returns True if at least one Node in the Model is trainable."""
    return any([n.trainable for n in self.lnodes])

  @trainable.setter
  def trainable(self, value: bool):
    """Freeze or unfreeze trainable Nodes in the Model."""
    for node in [n for n in self.lnodes]:
      node.trainable = value

  @property
  def lnodes(self) -> Tuple[Node]:
    return self._nodes

  @property
  def ff_edges(self) -> Sequence[Tuple[Node]]:
    return self._ff_edges

  @property
  def fb_edges(self) -> Sequence[Tuple[Node]]:
    return self._fb_edges

  @property
  def entry_nodes(self) -> Sequence[Node]:
    """First Nodes in the graph held by the Model."""
    return self._entry_nodes

  @property
  def exit_nodes(self) -> Sequence[Node]:
    """Last Nodes in the graph held by the Model."""
    return self._exit_nodes

  @property
  def feedback_nodes(self) -> Sequence[Node]:
    """Nodes which project feedback connections."""
    return tuple(self._fb_receivers.keys())

  @property
  def nodes_has_feedback(self) -> Sequence[Node]:
    """Nodes which receive feedback connections."""
    return tuple(self._fb_senders.keys())

  @property
  def ff_senders(self) -> Dict:
    """Nodes which project feedforward connections."""
    return self._ff_senders

  @property
  def ff_receivers(self) -> Dict:
    """Nodes which receive feedforward connections."""
    return self._ff_receivers

  @property
  def fb_senders(self) -> Dict:
    """Nodes which project feedback connections."""
    return self._fb_senders

  @property
  def fb_receivers(self) -> Dict:
    """Nodes which receive feedback connections."""
    return self._fb_receivers

  def update_graph(self,
                   new_nodes: Sequence[Node],
                   new_ff_edges: Sequence[Tuple[Node, Node]],
                   new_fb_edges: Sequence[Tuple[Node, Node]] = None) -> "Network":
    """Update current Model's with new nodes and edges, inplace (a copy
    is not performed).

    Parameters
    ----------
    new_nodes : list of Node
        New nodes.
    new_ff_edges : list of (Node, Node)
        New feedforward edges between nodes.
    new_fb_edges : list of (Node, Node)
        New feedback edges between nodes.

    Returns
    -------
    Network
        The updated network.
    """
    if new_fb_edges is None: new_fb_edges = tuple()
    self._nodes = tuple(set(new_nodes) | set(self.lnodes))
    self._ff_edges = tuple(set(new_ff_edges) | set(self.ff_edges))
    self._fb_edges = tuple(set(new_fb_edges) | set(self.fb_edges))
    # detect cycles in the graph flow
    if detect_cycle(self._nodes, self._ff_edges):
      raise ValueError('We detect cycles in feedforward connections. '
                       'Maybe you should replace some connection with '
                       'as feedback ones.')
    if detect_cycle(self._nodes, self._fb_edges):
      raise ValueError('We detect cycles in feedback connections. ')
    self._network_init()
    return self

  def replace_graph(self,
                    nodes: Sequence[Node],
                    ff_edges: Sequence[Tuple[Node, ...]],
                    fb_edges: Sequence[Tuple[Node, ...]] = None) -> "Network":
    if fb_edges is None: fb_edges = tuple()

    # assign nodes and edges
    self._nodes = tuple(nodes)
    self._ff_edges = tuple(ff_edges)
    self._fb_edges = tuple(fb_edges)
    self._network_init()
    return self

  def set_output_shape(self, shape: Dict[str, Sequence[int]]):
    # check shape
    if not isinstance(shape, dict):
      raise ValueError(f'Must be a dict of <node name, shape>, but got {type(shape)}: {shape}')
    for key, val in shape.items():
      if not isinstance(val, (tuple, list)):
        raise ValueError(f'Must be a sequence of int, but got {val} for key "{key}"')
      # for s in val:
      #   if not (isinstance(s, int) or (s is None)):
      #     raise ValueError(f'Must be a sequence of int, but got {val}')

    if not self._is_ff_initialized:
      if len(self.exit_nodes) == 1:
        self._output_shape = tuple(shape.values())[0]
      else:
        self._output_shape = shape
    else:
      for val in shape.values():
        check_shape_except_batch(val, self.output_shape)

  def init_ff_conn(self):
    """Initialize the feedforward connections of the network.
    This function will be called only once."""
    # input shapes of entry nodes
    for node in self.entry_nodes:
      # set ff shapes
      if node.feedforward_shapes is None:
        if self.feedforward_shapes is None:
          raise ValueError('Cannot find the input size. '
                           'Cannot initialize the network.')
        else:
          node.set_feedforward_shapes({node.name: self._feedforward_shapes[node.name]})
      # set fb shapes
      if node in self.fb_senders:
        fb_shapes = {node: node.output_shape for node in self.fb_senders.get(node, [])}
        if None not in fb_shapes.values():
          node.set_feedback_shapes(fb_shapes)
      # init ff conn
      node._init_ff_conn()

    # initialize the data
    children_queue = []
    ff_senders, _ = find_senders_and_receivers(self.ff_edges)

    # init shapes of other nodes
    for node in self._entry_nodes:
      for child in self.ff_receivers.get(node, []):
        ff_senders[child].remove(node)
        if len(ff_senders.get(child, [])) == 0:
          children_queue.append(child)
    while len(children_queue):
      node = children_queue.pop(0)
      # set ff shapes
      parent_sizes = {p: p.output_shape for p in self.ff_senders.get(node, [])}
      node.set_feedforward_shapes(parent_sizes)
      if node in self.fb_senders:
        # set fb shapes
        fb_shapes = {node: node.output_shape for node in self.fb_senders.get(node, [])}
        if None not in fb_shapes.values():
          node.set_feedback_shapes(fb_shapes)
      # init ff conn
      node._init_ff_conn()
      # append children
      for child in self.ff_receivers.get(node, []):
        ff_senders[child].remove(node)
        if len(ff_senders.get(child, [])) == 0:
          children_queue.append(child)

    # set output shape
    out_sizes = {node: node.output_shape for node in self.exit_nodes}
    self.set_output_shape(out_sizes)

  def init_fb_conn(self):
    """Initialize the feedback connections of the network.
    This function will be called only once."""
    for receiver, senders in self.fb_senders.items():
      fb_sizes = {node: node.output_shape for node in senders}
      if None in fb_sizes.values():
        none_size_nodes = [repr(n) for n, v in fb_sizes.items() if v is None]
        none_size_nodes = "\n".join(none_size_nodes)
        raise ValueError(f'Output shapes of nodes \n\n'
                         f'{none_size_nodes}\n\n'
                         f'have not been initialized, '
                         f'leading us cannot initialize the '
                         f'feedback connection of node \n\n'
                         f'{receiver}')
      receiver.set_feedback_shapes(fb_sizes)
      receiver._init_fb_conn()

  def _init_state(self, num_batch=1):
    """Initialize the states of all children nodes.
    This function can be called multiple times."""
    for node in self.lnodes:
      node._init_state(num_batch)
    self._is_state_initialized = True

  def _init_fb_output(self, num_batch=1):
    """Initialize the node feedback state.

    This function can be called multiple times. However,
    it is only triggered when the node has feedback connections.
    """
    for node in self.feedback_nodes:
      node._init_fb_output(num_batch)
    self._is_fb_state_initialized = True

  def initialize(self, num_batch: int = 1):
    """
    Initialize the whole network. This function must be called before applying JIT.

    This function is useful, because it is independent of the __call__ function.
    We can use this function before we apply JIT to __call__ function.
    """

    # set feedforward shapes
    if not self._is_ff_initialized:
      # check input and output nodes
      if len(self.entry_nodes) <= 0:
        raise ValueError(f"We found this network \n\n"
                         f"{self} "
                         f"\n\nhas no input nodes.")
      if len(self.exit_nodes) <= 0:
        raise ValueError(f"We found this network \n\n"
                         f"{self} "
                         f"\n\nhas no output nodes.")

      # check whether it has a feedforward path for each feedback pair
      ff_edges = [(a.name, b.name) for a, b in self.ff_edges]
      for node, receiver in self.fb_edges:
        if not detect_path(receiver.name, node.name, ff_edges):
          raise ValueError(f'Cannot build a feedback connection from '
                           f'\n\n{node} \n\n'
                           f'to '
                           f'\n\n{receiver} \n\n'
                           f'because there is no feedforward path between them. \n'
                           f'Maybe you should use "ff_connect" first to establish a '
                           f'feedforward connection between them. ')

      # feedforward checking
      in_sizes = dict()
      for node in self.entry_nodes:
        if node.feedforward_shapes is None:
          raise ValueError('Cannot initialize this node, because we detect '
                           '"feedforward_shapes" is None. '
                           'Maybe you need a brainpy.nn.Input instance '
                           'to instruct the input size.')
        in_sizes.update(node._feedforward_shapes)
      self.set_feedforward_shapes(in_sizes)

    # feedforward initialization
    if self.feedforward_shapes is None:
      raise ValueError('Cannot initialize this node, because we detect '
                       'both "feedforward_shapes" is None. ')
    check_integer(num_batch, 'num_batch', min_bound=1, allow_none=False)
    self._init_ff_conn()

    # initialize state
    self._init_state(num_batch)

    # set feedback shapes
    if not self._is_fb_initialized:
      if len(self.fb_senders) > 0:
        fb_sizes = dict()
        for sender in self.fb_senders.keys():
          fb_sizes[sender] = sender.output_shape
        self.set_feedback_shapes(fb_sizes)

    # feedback initialization
    if self.feedback_shapes is not None:
      self._init_fb_conn()

    # initialize feedback state
    self._init_fb_output(num_batch)

  def _check_inputs(self, ff, fb=None):
    # feedforward inputs
    if isinstance(ff, (bm.ndarray, jnp.ndarray)):
      ff = {self.entry_nodes[0].name: ff}
    if not isinstance(ff, dict):
      raise ValueError(f'ff must be a dict or a tensor, got {type(ff)}: {ff}')
    if len(self.entry_nodes) != len(ff):
      raise ValueError(f'This network has {len(self.entry_nodes)} '
                       f'entry nodes. While only {len(ff)} input '
                       f'data are given.')
    for n in self.entry_nodes:
      if n.name not in ff:
        raise ValueError(f'Cannot find the input of the node: \n{n}')
    for k, size in self._feedforward_shapes.items():
      if k not in ff:
        raise ValueError(f"The required key {k} is not provided in feedforward inputs.")
      if not check_shape_except_batch(size, ff[k].shape, mode='bool'):
        raise ValueError(f'Input size {ff[k].shape} is not consistent with '
                         f'the input size {size}')

    # feedback inputs
    if fb is not None:
      if isinstance(fb, (bm.ndarray, jnp.ndarray)):
        fb = {self.feedback_nodes[0]: fb}
      if not isinstance(fb, dict):
        raise ValueError(f'fb must be a dict or a tensor, '
                         f'got {type(fb)}: {fb}')
      if len(self.feedback_nodes) != len(fb):
        raise ValueError(f'This network has {len(self.feedback_nodes)} '
                         f'feedback nodes. While only {len(ff)} '
                         f'feedback data are given.')
      for n in self.feedback_nodes:
        if n.name not in fb:
          raise ValueError(f'Cannot find the feedback data from the node {n}')
      # check feedback consistency
      for k, size in self._feedback_shapes.items():
        if k not in fb:
          raise ValueError(f"The required key {k} is not provided in feedback inputs.")
        check_shape_except_batch(size, fb[k].shape)

    # data transformation
    ff = self.data_pass.filter(ff)
    fb = self.data_pass.filter(fb)
    return ff, fb

  def _call(self,
            ff: Union[Tensor, Dict[Any, Tensor]],
            fb: Optional[Union[Tensor, Dict[Any, Tensor]]] = None,
            forced_states: Optional[Dict[str, Tensor]] = None,
            forced_feedbacks: Optional[Dict[str, Tensor]] = None,
            monitors: Optional[Sequence[str]] = None,
            **kwargs):
    # initialization
    if not self.is_initialized:
      raise ValueError('Please initialize the Network first by calling "initialize()" function.')

    # initialize the forced data
    if forced_feedbacks is None: forced_feedbacks = dict()
    check_dict_data(forced_feedbacks, key_type=str, val_type=(bm.ndarray, jnp.ndarray))
    if forced_states is None: forced_states = dict()
    check_dict_data(forced_states, key_type=str, val_type=(bm.ndarray, jnp.ndarray))
    # initialize the monitors
    need_return_monitor = True
    if monitors is None:
      monitors = tuple()
      need_return_monitor = False
    attr_monitors: Dict[str, Tensor] = {}
    state_monitors: Dict[str, Tensor] = {}
    for key in monitors:
      if not isinstance(key, str):
        raise ValueError(f'"extra_returns" must be a sequence of string, '
                         f'while we got {type(key)}')
      splits = key.split('.')
      if len(splits) != 2:
        raise ValueError(f'Every term in "extra_returns" must be (node.item), '
                         f'while we got {key}')
      if splits[0] not in self.implicit_nodes:
        raise ValueError(f'Cannot found the node {splits[0]}, this network '
                         f'only has {list(self.implicit_nodes.keys())}.')

      if splits[1] not in NODE_STATES:  # attribute monitor
        if not hasattr(self.implicit_nodes[splits[0]], splits[1]):
          raise UnsupportedError(f'Each node can monitor its states (including {NODE_STATES}), '
                                 f'or its attribute. While {splits[1]} is neither the state nor '
                                 f'the attribute of node {splits[0]}.')
        else:
          attr_monitors[key] = getattr(self.implicit_nodes[splits[0]], splits[1])
      else:  # state monitor
        if splits[1] == 'state':
          assert self.implicit_nodes[splits[0]].state is not None, (f'{splits[0]} has no state, while '
                                                                    f'the user try to monitor it.')
        state_monitors[key] = None
    # calling the computation core
    ff, fb = self._check_inputs(ff, fb=fb)
    output, state_monitors = self.forward(ff, fb, forced_states, forced_feedbacks, state_monitors, **kwargs)
    if need_return_monitor:
      attr_monitors.update(state_monitors)
      return output, attr_monitors
    else:
      return output

  def _call_a_node(self, node, ff, fb, monitors, forced_states,
                   parent_outputs, children_queue, ff_senders,
                   **shared_kwargs):
    ff = node.data_pass.filter(ff)
    if f'{node.name}.inputs' in monitors:
      monitors[f'{node.name}.inputs'] = ff
    # get the output results
    if len(fb):
      fb = node.data_pass.filter(fb)
      if f'{node.name}.feedbacks' in monitors:
        monitors[f'{node.name}.feedbacks'] = fb
      parent_outputs[node] = node.forward(ff, fb, **shared_kwargs)
    else:
      parent_outputs[node] = node.forward(ff, **shared_kwargs)
    # get the feedback state
    if node in self.fb_receivers:
      node.set_fb_output(node.feedback(parent_outputs[node], **shared_kwargs))
    # forced state
    if node.name in forced_states:
      node.state.value = forced_states[node.name]
    # monitor the values
    if f'{node.name}.state' in monitors:
      monitors[f'{node.name}.state'] = node.state.value
    if f'{node.name}.output' in monitors:
      monitors[f'{node.name}.output'] = parent_outputs[node]
    # append children nodes
    for child in self.ff_receivers.get(node, []):
      ff_senders[child].remove(node)
      if len(ff_senders.get(child, [])) == 0:
        children_queue.append(child)

  def forward(self,
              ff,
              fb=None,
              forced_states: Dict[str, Tensor] = None,
              forced_feedbacks: Dict[str, Tensor] = None,
              monitors: Dict = None,
              **shared_kwargs):
    """The main computation function of a network.

    Parameters
    ----------
    ff: dict, sequence
      The feedforward inputs.
    fb: optional, dict, sequence
      The feedback inputs.
    forced_states: optional, dict
      The fixed state for the nodes in the network.
    forced_feedbacks: optional, dict
      The fixed feedback for the nodes in the network.
    monitors: optional, sequence
      Can be used to monitor the state or the attribute of a node in the network.
    **shared_kwargs
      Other parameters which will be parsed into every node.

    Returns
    -------
    Tensor
      A output tensor value, or a dict of output tensors.
    """
    all_nodes = set([n.name for n in self.lnodes])
    runned_nodes = set()
    output_nodes = set([n.name for n in self.exit_nodes])

    # initialize the feedback
    if forced_feedbacks is None: forced_feedbacks = dict()
    if monitors is None: monitors = dict()

    # initialize the data
    children_queue = []
    ff_senders, _ = find_senders_and_receivers(self.ff_edges)

    # initialize the parent output data
    parent_outputs = {}
    for i, node in enumerate(self._entry_nodes):
      ff_ = {node.name: ff[i]}
      fb_ = {p: (forced_feedbacks[p.name] if (p.name in forced_feedbacks) else p.fb_output)
             for p in self.fb_senders.get(node, [])}
      self._call_a_node(node, ff_, fb_, monitors, forced_states,
                        parent_outputs, children_queue, ff_senders,
                        **shared_kwargs)
      runned_nodes.add(node.name)

    # run the model
    while len(children_queue):
      node = children_queue.pop(0)
      # get feedforward and feedback inputs
      ff = {p: parent_outputs[p] for p in self.ff_senders.get(node, [])}
      fb = {p: (forced_feedbacks[p.name] if (p.name in forced_feedbacks) else p.fb_output)
            for p in self.fb_senders.get(node, [])}
      # call the node
      self._call_a_node(node, ff, fb, monitors, forced_states,
                        parent_outputs, children_queue, ff_senders,
                        **shared_kwargs)

      # - remove unnecessary parent outputs - #
      needed_parents = []
      runned_nodes.add(node.name)
      for child in (all_nodes - runned_nodes):
        for parent in self.ff_senders[self.implicit_nodes[child]]:
          needed_parents.append(parent.name)
      for parent in list(parent_outputs.keys()):
        _name = parent.name
        if _name not in needed_parents and _name not in output_nodes:
          parent_outputs.pop(parent)

    # returns
    if len(self.exit_nodes) > 1:
      state = {n.name: parent_outputs[n] for n in self.exit_nodes}
    else:
      state = parent_outputs[self.exit_nodes[0]]
    return state, monitors

  def plot_node_graph(self,
                      fig_size: tuple = (10, 10),
                      node_size: int = 2000,
                      arrow_size: int = 20,
                      layout='shell_layout'):
    """Plot the node graph based on NetworkX package

    Parameters
    ----------
    fig_size: tuple, default to (10, 10)
      The size of the figure
    node_size: int, default to 2000
      The size of the node
    arrow_size:int, default to 20
      The size of the arrow
    layout: str
      The graph layout. The supported layouts are:

      - "shell_layout"
      - "multipartite_layout"
      - "spring_layout"
      - "spiral_layout"
      - "spectral_layout"
      - "random_layout"
      - "planar_layout"
      - "kamada_kawai_layout"
      - "circular_layout"
    """
    try:
      import networkx as nx
    except (ModuleNotFoundError, ImportError):
      raise PackageMissingError('The node graph plotting currently need package "networkx". '
                                'But it can not be imported. ')
    try:
      import matplotlib.pyplot as plt
      from matplotlib.lines import Line2D
    except (ModuleNotFoundError, ImportError):
      raise PackageMissingError('The node graph plotting currently need package "matplotlib". '
                                'But it can not be imported. ')

    nodes_trainable = []
    nodes_untrainable = []
    for node in self.lnodes:
      if node.trainable:
        nodes_trainable.append(node.name)
      else:
        nodes_untrainable.append(node.name)

    ff_edges = []
    fb_edges = []
    rec_edges = []
    for edge in self.ff_edges:
      ff_edges.append((edge[0].name, edge[1].name))
    for edge in self.fb_edges:
      fb_edges.append((edge[0].name, edge[1].name))
    for node in self.lnodes:
      if isinstance(node, RecurrentNode):
        rec_edges.append((node.name, node.name))

    trainable_color = 'orange'
    untrainable_color = 'skyblue'
    ff_color = 'green'
    fb_color = 'red'
    rec_color = 'purple'
    G = nx.DiGraph()
    mid_nodes = list(set(self.lnodes) - set(self.entry_nodes) - set(self.exit_nodes))
    mid_nodes.sort(key=lambda x: x.name)
    index = 0
    for node in list(self.entry_nodes) + mid_nodes + list(self.exit_nodes):
      index = index + 1
      G.add_node(node.name, subset=index)
    G.add_edges_from(ff_edges)
    G.add_edges_from(fb_edges)
    G.add_edges_from(rec_edges)

    if layout not in SUPPORTED_LAYOUTS:
      raise UnsupportedError(f'Only support layouts: {SUPPORTED_LAYOUTS}')
    layout = getattr(nx, layout)(G)

    plt.figure(figsize=fig_size)
    nx.draw_networkx_nodes(G, pos=layout,
                           nodelist=nodes_trainable,
                           node_color=trainable_color,
                           node_size=node_size)
    nx.draw_networkx_nodes(G, pos=layout,
                           nodelist=nodes_untrainable,
                           node_color=untrainable_color,
                           node_size=node_size)

    ff_conn_style = "arc3,rad=0."
    nx.draw_networkx_edges(G, pos=layout,
                           edgelist=ff_edges,
                           edge_color=ff_color,
                           connectionstyle=ff_conn_style,
                           arrowsize=arrow_size,
                           node_size=node_size)
    fb_conn_style = "arc3,rad=0.3"
    nx.draw_networkx_edges(G, pos=layout,
                           edgelist=fb_edges,
                           edge_color=fb_color,
                           connectionstyle=fb_conn_style,
                           arrowsize=arrow_size,
                           node_size=node_size)
    rec_conn_style = "arc3,rad=-0.3"
    nx.draw_networkx_edges(G, pos=layout,
                           edgelist=rec_edges,
                           edge_color=rec_color,
                           arrowsize=arrow_size,
                           connectionstyle=rec_conn_style,
                           node_size=node_size,
                           node_shape='s')

    nx.draw_networkx_labels(G, pos=layout)
    proxie = []
    labels = []
    if len(nodes_trainable):
      proxie.append(Line2D([], [], color='white', marker='o',
                           markerfacecolor=trainable_color))
      labels.append('Trainable')
    if len(nodes_untrainable):
      proxie.append(Line2D([], [], color='white', marker='o',
                           markerfacecolor=untrainable_color))
      labels.append('Nontrainable')
    if len(ff_edges):
      proxie.append(Line2D([], [], color=ff_color, linewidth=2))
      labels.append('Feedforward')
    if len(fb_edges):
      proxie.append(Line2D([], [], color=fb_color, linewidth=2))
      labels.append('Feedback')
    if len(rec_edges):
      proxie.append(Line2D([], [], color=rec_color, linewidth=2))
      labels.append('Recurrent')

    plt.legend(proxie, labels, scatterpoints=1, markerscale=2, loc='best')
    plt.tight_layout()
    plt.show()


class FrozenNetwork(Network):
  """A FrozenNetwork is a Network that can not be linked to other nodes or networks."""

  def update_graph(self, new_nodes, new_ff_edges, new_fb_edges=None):
    raise TypeError(f"Cannot update FrozenModel {self}: "
                    f"model is frozen and cannot be modified.")

  def replace_graph(self, nodes, ff_edges, fb_edges=None):
    raise TypeError(f"Cannot update FrozenModel {self}: "
                    f"model is frozen and cannot be modified.")


class Sequential(Network):
  pass

# def _process_params(G, center, dim):
#     # Some boilerplate code.
#     import numpy as np
#
#     if not isinstance(G, nx.Graph):
#         empty_graph = nx.Graph()
#         empty_graph.add_nodes_from(G)
#         G = empty_graph
#
#     if center is None:
#         center = np.zeros(dim)
#     else:
#         center = np.asarray(center)
#
#     if len(center) != dim:
#         msg = "length of center coordinates must match dimension of layout"
#         raise ValueError(msg)
#
#     return G, center
#
#
# def multipartite_layout(G, subset_key="subset", align="vertical", scale=1, center=None):
#     import numpy as np
#
#     if align not in ("vertical", "horizontal"):
#       msg = "align must be either vertical or horizontal."
#       raise ValueError(msg)
#
#     G, center = _process_params(G, center=center, dim=2)
#     if len(G) == 0:
#       return {}
#
#     layers = {}
#     for v, data in G.nodes(data=True):
#       try:
#         layer = data[subset_key]
#       except KeyError:
#         msg = "all nodes must have subset_key (default='subset') as data"
#         raise ValueError(msg)
#       layers[layer] = [v] + layers.get(layer, [])
#
#     pos = None
#     nodes = []
#
#     width = len(layers)
#     for i, layer in layers.items():
#       height = len(layer)
#       xs = np.repeat(i, height)
#       ys = np.arange(0, height, dtype=float)
#       offset = ((width - 1) / 2, (height - 1) / 2)
#       layer_pos = np.column_stack([xs, ys]) - offset
#       if pos is None:
#         pos = layer_pos
#       else:
#         pos = np.concatenate([pos, layer_pos])
#       nodes.extend(layer)
#     pos = rescale_layout(pos, scale=scale) + center
#     if align == "horizontal":
#       pos = np.flip(pos, 1)
#     pos = dict(zip(nodes, pos))
#     return pos
#
#
# def rescale_layout(pos, scale=1):
#     """Returns scaled position array to (-scale, scale) in all axes.
#
#     The function acts on NumPy arrays which hold position information.
#     Each position is one row of the array. The dimension of the space
#     equals the number of columns. Each coordinate in one column.
#
#     To rescale, the mean (center) is subtracted from each axis separately.
#     Then all values are scaled so that the largest magnitude value
#     from all axes equals `scale` (thus, the aspect ratio is preserved).
#     The resulting NumPy Array is returned (order of rows unchanged).
#
#     Parameters
#     ----------
#     pos : numpy array
#         positions to be scaled. Each row is a position.
#
#     scale : number (default: 1)
#         The size of the resulting extent in all directions.
#
#     Returns
#     -------
#     pos : numpy array
#         scaled positions. Each row is a position.
#
#     See Also
#     --------
#     rescale_layout_dict
#     """
#     # Find max length over all dimensions
#     lim = 0  # max coordinate for all axes
#     for i in range(pos.shape[1]):
#         pos[:, i] -= pos[:, i].mean()
#         lim = max(abs(pos[:, i]).max(), lim)
#     # rescale to (-scale, scale) in all directions, preserves aspect
#     if lim > 0:
#         for i in range(pos.shape[1]):
#             pos[:, i] *= scale / lim
#     return pos



# -*- coding: utf-8 -*-

import numpy as np

from brainpy import math as bm
from brainpy.errors import MonitorError

__all__ = [
  'Monitor'
]


class Monitor(object):
  """The basic Monitor class to store the past variable trajectories.

  Currently, :py:class:`brainpy.simulation.Monitor` support to specify:

  - variable key by `strings`.
  - variable index by `None`, `int`, `list`, `tuple`, `1D array/tensor`
    (==> all will be transformed into a 1D array/tensor)
  - variable monitor interval by `None`, `int`, `float`

  Users can instance a monitor object by multiple ways:

  1. list of strings.

  >>> Monitor(variables=['a', 'b', 'c'])

  1.1. list of strings and list of intervals

  >>> Monitor(variables=['a', 'b', 'c'],
  >>>         intervals=[None, 1, 2] # ms
  >>>        )

  2. list of strings and string + indices

  >>> Monitor(variables=['a', ('b', bm.array([1,2,3])), 'c'])

  2.1. list of string (+ indices) and list of intervals

  >>> Monitor(variables=['a', ('b', bm.array([1,2,3])), 'c'],
  >>>         intervals=[None, 2, 3])

  3. a dictionary with the format of {key: indices}

  >>> Monitor(variables={'a': None, 'b': bm.array([1,2,3])})

  3.1. a dictionary of variable and indexes, and a dictionary of time intervals

  >>> Monitor(variables={'a': None, 'b': bm.array([1,2,3])},
  >>>         intervals={'b': 2.})

  .. note::
      :py:class:`brainpy.simulation.Monitor` records any target variable with an
      two-dimensional array/tensor with the shape of `(num_time_step, variable_size)`.
      This means for any variable, no matter what's the shape of the data
      (int, float, vector, matrix, 3D array/tensor), will be reshaped into a
      one-dimensional vector.

  """

  _KEYWORDS = ['_KEYWORDS', 'target', 'vars', 'intervals', 'ts', 'num_item',
               'item_names', 'item_indices', 'item_intervals', 'item_contents',
               'has_build']

  def __init__(self, variables, intervals=None):
    if isinstance(variables, (list, tuple)):
      if intervals is not None:
        if not isinstance(intervals, (list, tuple)):
          raise MonitorError(f'"vars" and "intervals" must be the same type. '
                             f'While we got type(vars)={type(variables)}, '
                             f'type(intervals)={type(intervals)}.')
        if len(variables) != len(intervals):
          raise MonitorError(f'The length of "vars" and "every" are not equal.')

    elif isinstance(variables, dict):
      if intervals is not None:
        if not isinstance(intervals, dict):
          raise MonitorError(f'"vars" and "every" must be the same type. '
                             f'While we got type(vars)={type(variables)}, '
                             f'type(intervals)={type(intervals)}.')
        for key in intervals.keys():
          if key not in variables:
            raise MonitorError(f'"{key}" is not in "vars": {list(variables.keys())}')

    else:
      raise MonitorError(f'We only supports a format of list/tuple/dict of '
                         f'"vars", while we got {type(variables)}.')

    self.has_build = False
    self.ts = None
    self.vars = variables
    self.intervals = intervals
    self.item_names = []
    self.item_indices = []
    self.item_intervals = []
    self.item_contents = dict()
    self.num_item = len(variables)
    self.build()

  def __repr__(self):
    return (f'{self.__class__.__name__}(items={tuple(self.item_names)}, '
            f'indices={self.item_indices})')

  def build(self):
    if not self.has_build:
      item_names = []
      item_indices = []
      item_contents = dict()

      if isinstance(self.vars, (list, tuple)):
        if self.intervals is None:
          item_intervals = [None] * len(self.vars)
        else:
          item_intervals = list(self.intervals)

        for mon_var, interval in zip(self.vars, item_intervals):
          # users monitor a variable by a string
          if isinstance(mon_var, str):
            mon_key = mon_var
            mon_idx = None
          # users monitor a variable by a tuple: `('b', bm.array([1,2,3]))`
          elif isinstance(mon_var, (tuple, list)):
            mon_key = mon_var[0]
            mon_idx = mon_var[1]
          else:
            raise MonitorError(f'Unknown monitor item: {str(mon_var)}')

          # self.check(mon_key)
          item_names.append(mon_key)
          item_indices.append(mon_idx)
          item_contents[mon_key] = []
          if interval is not None:
            item_contents[f'{mon_key}.t'] = []

      elif isinstance(self.vars, dict):
        item_intervals = []
        # users monitor a variable by a dict: `{'a': None, 'b': bm.array([1,2,3])}`
        for mon_key, mon_idx in self.vars.items():
          item_names.append(mon_key)
          item_indices.append(mon_idx)
          item_contents[mon_key] = []
          if self.intervals is None:
            item_intervals.append(None)
          else:
            if mon_key in self.intervals:
              item_intervals.append(self.intervals[mon_key])
            if self.intervals[mon_key] is not None:
              item_contents[f'{mon_key}.t'] = []
      else:
        raise MonitorError(f'Unknown monitors type: {type(self.vars)}')

      self.item_names = item_names
      self.item_indices = item_indices
      self.item_intervals = item_intervals
      self.item_contents = item_contents
      self.num_item = len(item_contents)
      self.has_build = True

  def __getitem__(self, item: str):
    """Get item in the monitor values.

    Parameters
    ----------
    item : str

    Returns
    -------
    value : ndarray
      The monitored values.
    """
    item_contents = super(Monitor, self).__getattribute__('item_contents')
    if item not in item_contents:
      raise ValueError(f'Do not have "{item}". Available items are:\n'
                       f'{list(item_contents.keys())}')
    return item_contents[item]

  def __setitem__(self, key, value):
    """Get item value in the monitor.

    Parameters
    ----------
    key : str
      The item key.
    value : ndarray
      The item value.
    """
    item_contents = super(Monitor, self).__getattribute__('item_contents')
    if key not in item_contents:
      raise ValueError(f'Do not have "{key}". Available items are:\n'
                       f'{list(item_contents.keys())}')
    self.item_contents[key] = value

  def __getattr__(self, item):
    if item in self._KEYWORDS:
      return super(Monitor, self).__getattribute__(item)
    else:
      item_contents = super(Monitor, self).__getattribute__('item_contents')
      if item in item_contents:
        return item_contents[item]
      else:
        super(Monitor, self).__getattribute__(item)

  def __setattr__(self, key, value):
    if key in self._KEYWORDS:
      object.__setattr__(self, key, value)
    elif key in self.item_contents:
      self.item_contents[key] = value
    else:
      object.__setattr__(self, key, value)

  def numpy(self):
    for key, val in self.item_contents.items():
      self.item_contents[key] = np.asarray(val)
    if self.ts is not None:
      self.ts = np.asarray(self.ts)



# -*- coding: utf-8 -*-

r"""This module provides exponential integrators for ODEs.

Exponential integrators are a large class of methods from numerical analysis is based on
the exact integration of the linear part of the initial value problem. Because the linear
part is integrated exactly, this can help to mitigate the stiffness of a differential
equation.

We consider initial value problems of the form,

.. math:: u'(t)=f(u(t)),\qquad u(t_{0})=u_{0},

which can be decomposed of

.. math:: u'(t)=Lu(t)+N(u(t)),\qquad u(t_{0})=u_{0},

where :math:`L={\frac {\partial f}{\partial u}}` (the Jacobian of f) is composed of
linear terms, and :math:`N=f(u)-Lu` is composed of the non-linear terms.

This procedure enjoys the advantage, in each step, that
:math:`{\frac {\partial N_{n}}{\partial u}}(u_{n})=0`.
This considerably simplifies the derivation of the order conditions and improves the
stability when integrating the nonlinearity :math:`N(u(t))`.

Exact integration of this problem from time 0 to a later time :math:`t` can be performed
using `matrix exponentials <https://en.wikipedia.org/wiki/Matrix_exponential>`_ to define
an integral equation for the exact solution:

.. math:: u(t)=e^{Lt}u_{0}+\int _{0}^{t}e^{L(t-\tau )}N\left(t+\tau, u\left(\tau \right)\right)\,d\tau .

This representation of the exact solution is also called as *variation-of-constant formula*.
In the case of :math:`N\equiv 0`, this formulation is the exact solution to the linear
differential equation.


**Exponential Rosenbrock methods**

Exponential Rosenbrock methods were shown to be very efficient in solving large systems
of stiff ODEs. Applying the variation-of-constants formula gives the exact solution at
time :math:`t_{n+1}` with the numerical solution :math:`u_n` as

.. math::
    u(t_{n+1})=e^{h_{n}L}u(t_{n})+\int _{0}^{h_{n}}e^{(h_{n}-\tau )L}N(t_n+\tau, u(t_{n}+\tau ))d\tau .
    :label: discrete-variation-of-constants-formula

where :math:`h_n=t_{n+1}-t_n`.

The idea now is to approximate the integral in :eq:`discrete-variation-of-constants-formula`
by some quadrature rule with nodes :math:`c_{i}` and weights :math:`b_{i}(h_{n}L)`
(:math:`1\leq i\leq s`). This yields the following class of *s-stage* explicit exponential
Rosenbrock methods:

.. math::
    \begin{align}
    U_{ni}=&e^{c_{i}h_{n}L}u_n+h_{n}\sum_{j=1}^{i-1}a_{ij}(h_{n}L)N(U_{nj}),  \\
    u_{n+1}=&e^{h_{n}L}u_n+h_{n}\sum_{i=1}^{s}b_{i}(h_{n}L)N(U_{ni})
    \end{align}

where :math:`U_{ni}\approx u(t_{n}+c_{i}h_{n})`.

The coefficients :math:`a_{ij}(z),b_{i}(z)` are usually chosen as linear combinations of
the entire functions :math:`\varphi _{k}(c_{i}z),\varphi _{k}(z)`, respectively, where

.. math::
    \begin{align}
    \varphi _{k}(z)=&\int _{0}^{1}e^{(1-\theta )z}{\frac {\theta ^{k-1}}{(k-1)!}}d\theta ,\quad k\geq 1, \\
    \varphi _{0}(z)=&e^{z},\\
    \varphi _{k+1}(z)=&{\frac {\varphi_{k}(z)-\varphi _{k}(0)}{z}},\ k\geq 0.
    \end{align}

By introducing the difference :math:`D_{ni}=N(U_{ni})-N(u_{n})`, they can be reformulated
in a more efficient way for implementation as

.. math::
    \begin{align}
    U_{ni}=&u_{n}+c_{i}h_{n}\varphi _{1}(c_{i}h_{n}L)f(u_{n})+h_{n}\sum _{j=2}^{i-1}a_{ij}(h_{n}L)D_{nj}, \\
    u_{n+1}=&u_{n}+h_{n}\varphi _{1}(h_{n}L)f(u_{n})+h_{n}\sum _{i=2}^{s}b_{i}(h_{n}L)D_{ni}.
    \end{align}

where :math:`\varphi_{1}(z)=\frac{e^z-1}{z}`.

In order to implement this scheme with adaptive step size, one can consider, for the purpose
of local error estimation, the following embedded methods

.. math:: {\bar {u}}_{n+1}=u_{n}+h_{n}\varphi _{1}(h_{n}L)f(u_{n})+h_{n}\sum _{i=2}^{s}{\bar {b}}_{i}(h_{n}L)D_{ni},

which use the same stages :math:`U_{ni}` but with weights :math:`{\bar {b}}_{i}`.

For convenience, the coefficients of the explicit exponential Rosenbrock methods together
with their embedded methods can be represented by using the so-called reduced Butcher
tableau as follows:

.. math::
    \begin{array}{c|ccccc}
    c_{2} & & & & & \\
    c_{3} & a_{32} & & & & \\
    \vdots & \vdots & & \ddots & & \\
    c_{s} & a_{s 2} & a_{s 3} & \cdots & a_{s, s-1} \\
    \hline & b_{2} & b_{3} & \cdots & b_{s-1} & b_{s} \\
    & \bar{b}_{2} & \bar{b}_{3} & \cdots & \bar{b}_{s-1} & \bar{b}_{s}
    \end{array}

.. [1] https://en.wikipedia.org/wiki/Exponential_integrator
.. [2] Hochbruck, M., & Ostermann, A. (2010). Exponential integrators. Acta Numerica, 19, 209-286.
"""

import inspect
import logging

from brainpy import math, errors, tools
from brainpy.base.collector import Collector
from brainpy.integrators import constants as C, utils, joint_eq
from brainpy.integrators.analysis_by_ast import separate_variables
from brainpy.integrators.ode.base import ODEIntegrator
from .generic import register_ode_integrator

try:
  import sympy
  from brainpy.integrators import analysis_by_sympy
except (ModuleNotFoundError, ImportError):
  sympy = analysis_by_sympy = None

logger = logging.getLogger('brainpy.integrators.ode.exponential')

__all__ = [
  'ExponentialEuler',
  'ExpEulerAuto',
]


class ExponentialEuler(ODEIntegrator):
  r"""The exponential Euler method for ODEs.

  The simplest exponential Rosenbrock method is the exponential
  Rosenbrock–Euler scheme, which has order 2.

  For an ODE equation of the form

  .. math::

      u^{\prime}=f(u), \quad u(0)=u_{0}

  its schema is given by

  .. math::

      u_{n+1}= u_{n}+h \varphi(hL) f (u_{n})

  where :math:`L=f^{\prime}(u_{n})` and :math:`\varphi(z)=\frac{e^{z}-1}{z}`.

  For a linear ODE system: :math:`u^{\prime} = Ay + B`,
  the above equation is equal to :math:`u_{n+1}= u_{n}e^{hA}-B/A(1-e^{hA})`,
  which is the exact solution for this ODE system.

  Examples
  --------

  Linear system example: HH model's derivative function.

  >>> def derivative(V, m, h, n, t, Iext, gNa, ENa, gK, EK, gL, EL, C):
  >>>    alpha = 0.1 * (V + 40) / (1 - bm.exp(-(V + 40) / 10))
  >>>    beta = 4.0 * bm.exp(-(V + 65) / 18)
  >>>    dmdt = alpha * (1 - m) - beta * m
  >>>
  >>>    alpha = 0.07 * bm.exp(-(V + 65) / 20.)
  >>>    beta = 1 / (1 + bm.exp(-(V + 35) / 10))
  >>>    dhdt = alpha * (1 - h) - beta * h
  >>>
  >>>    alpha = 0.01 * (V + 55) / (1 - bm.exp(-(V + 55) / 10))
  >>>    beta = 0.125 * bm.exp(-(V + 65) / 80)
  >>>    dndt = alpha * (1 - n) - beta * n
  >>>
  >>>    I_Na = (gNa * m ** 3.0 * h) * (V - ENa)
  >>>    I_K = (gK * n ** 4.0) * (V - EK)
  >>>    I_leak = gL * (V - EL)
  >>>    dVdt = (- I_Na - I_K - I_leak + Iext) / C
  >>>    return dVdt, dmdt, dhdt, dndt
  >>>
  >>> ExponentialEuler(f=derivative, show_code=True)
  def brainpy_itg_of_ode0_drivative(V, m, h, n, t, Iext, gNa, ENa, gK, EK, gL, EL, C, dt=0.01):
    _dfV_dt = EK*gK*n**4.0/C + EL*gL/C + ENa*gNa*h*m**3.0/C + Iext/C - V*gK*n**4.0/C - V*gL/C - V*gNa*h*m**3.0/C
    _V_linear = -gK*n**4.0/C - gL/C - gNa*h*m**3.0/C
    _V_linear_exp = math.exp(_V_linear * dt)
    _V_df_part = _dfV_dt*(_V_linear_exp - 1)/_V_linear
    V_new = V + _V_df_part
    #
    alpha = 0.1 * (V + 40) / (1 - bm.exp(-(V + 40) / 10))
    beta = 4.0 * bm.exp(-(V + 65) / 18)
    _dfm_dt = -alpha*m + 1.0*alpha - beta*m
    _m_linear = -alpha - beta
    _m_linear_exp = math.exp(_m_linear * dt)
    _m_df_part = _dfm_dt*(_m_linear_exp - 1)/_m_linear
    m_new = _m_df_part + m
    #
    alpha = 0.07 * bm.exp(-(V + 65) / 20.0)
    beta = 1 / (1 + bm.exp(-(V + 35) / 10))
    _dfh_dt = -alpha*h + 1.0*alpha - beta*h
    _h_linear = -alpha - beta
    _h_linear_exp = math.exp(_h_linear * dt)
    _h_df_part = _dfh_dt*(_h_linear_exp - 1)/_h_linear
    h_new = _h_df_part + h
    #
    alpha = 0.01 * (V + 55) / (1 - bm.exp(-(V + 55) / 10))
    beta = 0.125 * bm.exp(-(V + 65) / 80)
    _dfn_dt = -alpha*n + 1.0*alpha - beta*n
    _n_linear = -alpha - beta
    _n_linear_exp = math.exp(_n_linear * dt)
    _n_df_part = _dfn_dt*(_n_linear_exp - 1)/_n_linear
    n_new = _n_df_part + n
    #
    return V_new, m_new, h_new, n_new

  Nonlinear system example: Van der Pol oscillator.

  >>> def vdp_derivative(x, y, t, mu):
  >>>    dx = mu * (x - x ** 3 / 3 - y)
  >>>    dy = x / mu
  >>>    return dx, dy
  >>>
  >>> ExponentialEuler(f=vdp_derivative, show_code=True)
  def brainpy_itg_of_ode0_vdp_derivative(x, y, t, mu, dt=0.01):
    _dfx_dt = mu*x - 0.333333333333333*mu*x**3.0 - mu*y
    _x_linear = -0.999999999999999*mu*x**2.0 + mu
    _x_linear_exp = math.exp(_x_linear * dt)
    _x_df_part = _dfx_dt*(_x_linear_exp - 1)/_x_linear
    x_new = _x_df_part + x
    #
    dy = x / mu
    _dfy_dt = dy
    _y_df_part = _dfy_dt * dt
    y_new = _y_df_part + y
    #
    return x_new, y_new

  However, ExponentialEuler method has severve constraints (see below).

  .. note::
    Many constraints are involved when using this ExponentialEuler method, because
    it uses SymPy to make symbolic differentiation for your codes.
    If you want to use get a more flexible method to do exponential euler integration,
    please refer to ``ExpEulerAuto`` method.

    The mechanism of ExponentialEuler method is :
    First, the user's codes are transformed into the SymPy expressions
    by using AST. Then, infer the derivative of the sympy expression by
    using ``sympy.diff()``.

    The constains of using ExponentialEuler method are:

    1. Must use latex-like codes to write equations.

    For your derivative variable, if there are general Python functions are using
    these varaibles, the SymPy parsing will not work.
    For instance, the following codes can not be recognized by SymPy, because the
    unsupported function ``bm.power`` are performing on the derivative variable ``x``.

    >>> import brainpy as bp
    >>> import brainpy.math as bm
    >>>
    >>> def derivative(x, t):
    >>>   dx = bm.power(x, 3) - bm.power(x, 2)
    >>>   return dx
    >>>
    >>> f = bp.ode.ExponentialEuler(derivative, show_code=True)
    def brainpy_itg_of_ode1_dev(x, t, dt=0.1):
      _dfx_dt = -bm.power(x, 2.0) + bm.power(x, 3.0)
      _x_linear = -Derivative(bm.power(x, 2.0), x) + Derivative(bm.power(x, 3.0), x)
      _x_linear_exp = math.exp(_x_linear * dt)
      _x_df_part = _dfx_dt*(_x_linear_exp - 1)/_x_linear
      x_new = _x_df_part + x
      return x_new

    As you see, SymPy cannot parse ``bm.power()``. Therefore, the linear part
    (``_x_linear``) of the system contains a unknown variable ``Derivative``. This will
    cause an error in future. Instead, you must define the derivative as:

    >>> def derivative(x, t):
    >>>   dx = x ** 3 - x ** 2
    >>>   return dx
    >>>
    >>> f = bp.ode.ExponentialEuler(derivative, show_code=True)
    def brainpy_itg_of_ode3_dev(x, t, dt=0.1):
      _dfx_dt = -x**2.0 + x**3.0
      _x_linear = -2.0*x**1.0 + 3.0*x**2.0
      _x_linear_exp = math.exp(_x_linear * dt)
      _x_df_part = _dfx_dt*(_x_linear_exp - 1)/_x_linear
      x_new = _x_df_part + x
      return x_new

    So, what ``brainpy.math`` functions are supported on derivative variables?
    You can inspect them by:

    >>> bp.integrators.analysis_by_sympy.FUNCTION_MAPPING.keys()
    {'abs': Abs,
     'sign': sign,
     'sinc': sinc,
     'arcsin': asin,
     'arccos': acos,
     'arctan': atan,
     'arctan2': atan2,
     'arcsinh': asinh,
     'arccosh': acosh,
     'arctanh': atanh,
     'log2': log2,
     'log1p': log1p,
     'expm1': expm1,
     'exp2': exp2,
     'asin': asin,
     'acos': acos,
     'atan': atan,
     'atan2': atan2,
     'asinh': asinh,
     'acosh': acosh,
     'atanh': atanh,
     'cos': cos,
     'sin': sin,
     'tan': tan,
     'cosh': cosh,
     'sinh': sinh,
     'tanh': tanh,
     'log': log,
     'log10': log10,
     'sqrt': sqrt,
     'exp': exp,
     'hypot': hypot,
     'ceil': ceiling,
     'floor': floor}

    However, if your general Python functions are used in variables not related to
    the derivative variable, it will work. For instance,

    >>> def derivative(x, t, y):
    >>>   alpha = bm.power(y, 3) - bm.power(y, 2)
    >>>   beta = bm.cumsum(y)
    >>>   dx = alpha * x - beta * (1 - x)
    >>>   return dx
    >>>
    >>> bp.ode.ExponentialEuler(derivative, show_code=True)

    2. Functional return only support symbols, not expressions.

    For example, this derivative will cause an error:

    >>> def derivative(x, t, tau):
    >>>   return -x / tau
    >>>
    >>> bp.ode.ExponentialEuler(derivative)
    brainpy.errors.DiffEqError: Cannot analyze differential equation with expression return. Like:
    def df(v, t):
        return -v + 1.
    We only support return variables. Therefore, the above should be coded as:
    def df(v, t):
        dv = -v + 1.
        return dv

    Instead, you should wrap the expression ``-x / tau`` as a symbol ``dx``, then return ``dx``.

    >>> def derivative(x, t, tau):
    >>>   dx = -x / tau
    >>>   return dx
    >>>
    >>> bp.ode.ExponentialEuler(derivative)
    <brainpy.integrators.ode.exponential.ExponentialEuler at 0x29daa298d60>

  See Also
  --------
  ExpEulerAuto

  Parameters
  ----------
  f : function
    The derivative function.
  dt : optional, float
    The numerical precision.
  var_type : optional, str
    The variable type.
  show_code : bool
    Whether show the code.
  timeout : float
    The timeout limit to use sympy solver.
  """

  def __init__(self,
               f,
               var_type=None,
               dt=None,
               name=None,
               show_code=False,
               timeout=5,
               state_delays=None,
               neutral_delays=None):
    super(ExponentialEuler, self).__init__(f=f,
                                           var_type=var_type,
                                           dt=dt,
                                           name=name,
                                           show_code=show_code,
                                           state_delays=state_delays,
                                           neutral_delays=neutral_delays)

    self.timeout = timeout

    # keyword checking
    keywords = {
      C.F: 'the derivative function',
      # C.DT: 'the precision of numerical integration',
      'exp': 'the exponential function',
      'math': 'the math module',
    }
    for v in self.variables:
      keywords[f'{v}_new'] = 'the intermediate value'
    utils.check_kws(self.arg_names, keywords)

    # build the integrator
    self.build()

  def build(self):
    if analysis_by_sympy is None or sympy is None:
      raise errors.PackageMissingError(f'Package "sympy" must be installed when the users '
                                       f'want to utilize {ExponentialEuler.__name__}. ')

    # check bound method
    if hasattr(self.f, '__self__'):
      self.code_lines = [f'def {self.func_name}({", ".join(["self"] + list(self.arguments))}):']

    # code scope
    closure_vars = inspect.getclosurevars(self.f)
    self.code_scope.update(closure_vars.nonlocals)
    self.code_scope.update(dict(closure_vars.globals))
    self.code_scope['math'] = math

    analysis = separate_variables(self.f)
    variables_for_returns = analysis['variables_for_returns']
    expressions_for_returns = analysis['expressions_for_returns']
    for vi, (key, all_var) in enumerate(variables_for_returns.items()):
      # separate variables
      sd_variables = []
      for v in all_var:
        if len(v) > 1:
          raise ValueError(f'Cannot analyze multi-assignment code line: {v}.')
        sd_variables.append(v[0])
      expressions = expressions_for_returns[key]
      var_name = self.variables[vi]
      diff_eq = analysis_by_sympy.SingleDiffEq(var_name=var_name,
                                               variables=sd_variables,
                                               expressions=expressions,
                                               derivative_expr=key,
                                               scope=self.code_scope,
                                               func_name=self.func_name)
      var = sympy.Symbol(diff_eq.var_name, real=True)
      try:
        s_df_part = tools.timeout(self.timeout)(self.solve)(diff_eq, var)
      except KeyboardInterrupt:
        raise errors.DiffEqError(
          f'{self.__class__} solve {self.f} failed, because '
          f'symbolic differentiation of SymPy timeout due to {self.timeout} s limit. '
          f'Instead, you can use {ExpEulerAuto} to make Exponential Euler '
          f'integration due to due to it is capable of '
          f'performing automatic differentiation.'
        )
      # update expression
      update = var + s_df_part

      # The actual update step
      self.code_lines.append(f'  {diff_eq.var_name}_new = {analysis_by_sympy.sympy2str(update)}')
      self.code_lines.append('')

    self.code_lines.append(f'  return {", ".join([f"{v}_new" for v in self.variables])}')
    self.integral = utils.compile_code(code_scope={k: v for k, v in self.code_scope.items()},
                                       code_lines=self.code_lines,
                                       show_code=self.show_code,
                                       func_name=self.func_name)

    if hasattr(self.f, '__self__'):
      host = self.f.__self__
      self.integral = self.integral.__get__(host, host.__class__)

  def solve(self, diff_eq, var):
    if analysis_by_sympy is None or sympy is None:
      raise errors.PackageMissingError(f'Package "sympy" must be installed when the users '
                                       f'want to utilize {ExponentialEuler.__name__}. ')

    f_expressions = diff_eq.get_f_expressions(substitute_vars=diff_eq.var_name)

    # code lines
    self.code_lines.extend([f"  {str(expr)}" for expr in f_expressions[:-1]])

    # get the linear system using sympy
    f_res = f_expressions[-1]
    if len(f_res.code) > 500:
      raise errors.DiffEqError(
        f'Too complex differential equation:\n\n'
        f'{f_res.code}\n\n'
        f'SymPy cannot analyze. Please use {ExpEulerAuto} to '
        f'make Exponential Euler integration due to it is capable of '
        f'performing automatic differentiation.'
      )
    df_expr = analysis_by_sympy.str2sympy(f_res.code).expr.expand()
    s_df = sympy.Symbol(f"{f_res.var_name}")
    self.code_lines.append(f'  {s_df.name} = {analysis_by_sympy.sympy2str(df_expr)}')

    # get df part
    s_linear = sympy.Symbol(f'_{diff_eq.var_name}_linear')
    s_linear_exp = sympy.Symbol(f'_{diff_eq.var_name}_linear_exp')
    s_df_part = sympy.Symbol(f'_{diff_eq.var_name}_df_part')
    if df_expr.has(var):
      # linear
      linear = sympy.diff(df_expr, var, evaluate=True)
      # TODO: linear has unknown symbol
      self.code_lines.append(f'  {s_linear.name} = {analysis_by_sympy.sympy2str(linear)}')
      # linear exponential
      self.code_lines.append(f'  {s_linear_exp.name} = math.exp({s_linear.name} * {C.DT})')
      # df part
      df_part = (s_linear_exp - 1) / s_linear * s_df
      self.code_lines.append(f'  {s_df_part.name} = {analysis_by_sympy.sympy2str(df_part)}')
    else:
      # df part
      self.code_lines.append(f'  {s_df_part.name} = {s_df.name} * {C.DT}')
    return s_df_part


register_ode_integrator('exponential_euler', ExponentialEuler)
register_ode_integrator('exp_euler', ExponentialEuler)


class ExpEulerAuto(ODEIntegrator):
  """Exponential Euler method using automatic differentiation.

  This method uses `brainpy.math.vector_grad <../../math/generated/brainpy.math.autograd.vector_grad.html>`_
  to automatically infer the linear part of the given function. Therefore, it has minimal constraints
  on your derivative function. Arbitrary complex functions can be numerically integrated with this method.

  Examples
  --------

  Here is an example uses ``ExpEulerAuto`` to implement HH neuron model.

  .. plot::
    :include-source: True

    >>> import brainpy as bp
    >>> import brainpy.math as bm
    >>>
    >>> class HH(bp.dyn.NeuGroup):
    >>>   def __init__(self, size, ENa=55., EK=-90., EL=-65, C=1.0, gNa=35., gK=9.,
    >>>                gL=0.1, V_th=20., phi=5.0, name=None):
    >>>     super(HH, self).__init__(size=size, name=name)
    >>>
    >>>     # parameters
    >>>     self.ENa = ENa
    >>>     self.EK = EK
    >>>     self.EL = EL
    >>>     self.C = C
    >>>     self.gNa = gNa
    >>>     self.gK = gK
    >>>     self.gL = gL
    >>>     self.V_th = V_th
    >>>     self.phi = phi
    >>>
    >>>     # variables
    >>>     self.V = bm.Variable(bm.ones(size) * -65.)
    >>>     self.h = bm.Variable(bm.ones(size) * 0.6)
    >>>     self.n = bm.Variable(bm.ones(size) * 0.32)
    >>>     self.spike = bm.Variable(bm.zeros(size, dtype=bool))
    >>>     self.input = bm.Variable(bm.zeros(size))
    >>>
    >>>     # functions
    >>>     self.int_h = bp.ode.ExpEulerAuto(self.dh)
    >>>     self.int_n = bp.ode.ExpEulerAuto(self.dn)
    >>>     self.int_V = bp.ode.ExpEulerAuto(self.dV)
    >>>
    >>>   def dh(self, h, t, V):
    >>>     alpha = 0.07 * bm.exp(-(V + 58) / 20)
    >>>     beta = 1 / (bm.exp(-0.1 * (V + 28)) + 1)
    >>>     dhdt = self.phi * (alpha * (1 - h) - beta * h)
    >>>     return dhdt
    >>>
    >>>   def dn(self, n, t, V):
    >>>     alpha = -0.01 * (V + 34) / (bm.exp(-0.1 * (V + 34)) - 1)
    >>>     beta = 0.125 * bm.exp(-(V + 44) / 80)
    >>>     dndt = self.phi * (alpha * (1 - n) - beta * n)
    >>>     return dndt
    >>>
    >>>   def dV(self, V, t, h, n, Iext):
    >>>     m_alpha = -0.1 * (V + 35) / (bm.exp(-0.1 * (V + 35)) - 1)
    >>>     m_beta = 4 * bm.exp(-(V + 60) / 18)
    >>>     m = m_alpha / (m_alpha + m_beta)
    >>>     INa = self.gNa * m ** 3 * h * (V - self.ENa)
    >>>     IK = self.gK * n ** 4 * (V - self.EK)
    >>>     IL = self.gL * (V - self.EL)
    >>>     dVdt = (- INa - IK - IL + Iext) / self.C
    >>>
    >>>     return dVdt
    >>>
    >>>   def update(self, _t, _dt):
    >>>     h = self.int_h(self.h, _t, self.V, dt=_dt)
    >>>     n = self.int_n(self.n, _t, self.V, dt=_dt)
    >>>     V = self.int_V(self.V, _t,  self.h, self.n, self.input, dt=_dt)
    >>>     self.spike.value = bm.logical_and(self.V < self.V_th, V >= self.V_th)
    >>>     self.V.value = V
    >>>     self.h.value = h
    >>>     self.n.value = n
    >>>     self.input[:] = 0.
    >>>
    >>> run = bp.dyn.DSRunner(HH(1), inputs=('input', 2.), monitors=['V'], dt=0.05)
    >>> run(100)
    >>> bp.visualize.line_plot(run.mon.ts, run.mon.V, legend='V', show=True)

  The above example can also be defined with ``brainpy.JointEq``.

  .. plot::
    :include-source: True

    >>> import brainpy as bp
    >>> import brainpy.math as bm
    >>>
    >>> class HH(bp.dyn.NeuGroup):
    >>>   def __init__(self, size, ENa=55., EK=-90., EL=-65, C=1.0, gNa=35., gK=9.,
    >>>                gL=0.1, V_th=20., phi=5.0, name=None):
    >>>     super(HH, self).__init__(size=size, name=name)
    >>>
    >>>     # parameters
    >>>     self.ENa = ENa
    >>>     self.EK = EK
    >>>     self.EL = EL
    >>>     self.C = C
    >>>     self.gNa = gNa
    >>>     self.gK = gK
    >>>     self.gL = gL
    >>>     self.V_th = V_th
    >>>     self.phi = phi
    >>>
    >>>     # variables
    >>>     self.V = bm.Variable(bm.ones(size) * -65.)
    >>>     self.h = bm.Variable(bm.ones(size) * 0.6)
    >>>     self.n = bm.Variable(bm.ones(size) * 0.32)
    >>>     self.spike = bm.Variable(bm.zeros(size, dtype=bool))
    >>>     self.input = bm.Variable(bm.zeros(size))
    >>>
    >>>     # functions
    >>>     derivative = bp.JointEq([self.dh, self.dn, self.dV])
    >>>     self.integral = bp.ode.ExpEulerAuto(derivative)
    >>>
    >>>   def dh(self, h, t, V):
    >>>     alpha = 0.07 * bm.exp(-(V + 58) / 20)
    >>>     beta = 1 / (bm.exp(-0.1 * (V + 28)) + 1)
    >>>     dhdt = self.phi * (alpha * (1 - h) - beta * h)
    >>>     return dhdt
    >>>
    >>>   def dn(self, n, t, V):
    >>>     alpha = -0.01 * (V + 34) / (bm.exp(-0.1 * (V + 34)) - 1)
    >>>     beta = 0.125 * bm.exp(-(V + 44) / 80)
    >>>     dndt = self.phi * (alpha * (1 - n) - beta * n)
    >>>     return dndt
    >>>
    >>>   def dV(self, V, t, h, n, Iext):
    >>>     m_alpha = -0.1 * (V + 35) / (bm.exp(-0.1 * (V + 35)) - 1)
    >>>     m_beta = 4 * bm.exp(-(V + 60) / 18)
    >>>     m = m_alpha / (m_alpha + m_beta)
    >>>     INa = self.gNa * m ** 3 * h * (V - self.ENa)
    >>>     IK = self.gK * n ** 4 * (V - self.EK)
    >>>     IL = self.gL * (V - self.EL)
    >>>     dVdt = (- INa - IK - IL + Iext) / self.C
    >>>
    >>>     return dVdt
    >>>
    >>>   def update(self, _t, _dt):
    >>>     h, n, V = self.integral(self.h, self.n, self.V, _t, self.input, dt=_dt)
    >>>     self.spike.value = bm.logical_and(self.V < self.V_th, V >= self.V_th)
    >>>     self.V.value = V
    >>>     self.h.value = h
    >>>     self.n.value = n
    >>>     self.input[:] = 0.
    >>>
    >>> run = bp.dyn.DSRunner(HH(1), inputs=('input', 2.), monitors=['V'], dt=0.05)
    >>> run(100)
    >>> bp.visualize.line_plot(run.mon.ts, run.mon.V, legend='V', show=True)

  See Also
  --------
  ExponentialEuler

  Parameters
  ----------
  f : function, joint_eq.JointEq
    The derivative function.
  var_type : optional, str
    The variable type.
  dt : optional, float
    The default numerical integration step.
  name : optional, str
    The integrator name.
  show_code : bool
  dyn_var : optional, dict, sequence of JaxArray, JaxArray
  """

  def __init__(self,
               f,
               var_type=None,
               dt=None,
               name=None,
               show_code=False,
               dyn_var=None,
               state_delays=None,
               neutral_delays=None):
    super(ExpEulerAuto, self).__init__(f=f,
                                       var_type=var_type,
                                       dt=dt,
                                       name=name,
                                       show_code=show_code,
                                       state_delays=state_delays,
                                       neutral_delays=neutral_delays)

    self.dyn_var = dyn_var

    # keyword checking
    keywords = {
      C.F: 'the derivative function',
      # C.DT: 'the precision of numerical integration',
    }
    utils.check_kws(self.arg_names, keywords)

    # build the integrator
    self.code_lines = []
    self.code_scope = {}
    self.integral = self.build()

  def build(self):
    all_vars, all_pars = [], []
    integrals, arg_names = [], []
    a = self._build_integrator(self.f)
    for integral, vars, _ in a:
      integrals.append(integral)
      for var in vars:
        if var not in all_vars:
          all_vars.append(var)
    for _, vars, pars in a:
      for par in pars:
        if (par not in all_vars) and (par not in all_pars):
          all_pars.append(par)
      arg_names.append(vars + pars + ['dt'])
    all_pars.append('dt')
    all_vps = all_vars + all_pars

    def integral_func(*args, **kwargs):
      # format arguments
      params_in = Collector()
      for i, arg in enumerate(args):
        params_in[all_vps[i]] = arg
      params_in.update(kwargs)
      if 'dt' not in params_in:
        params_in['dt'] = math.get_dt()

      # call integrals
      results = []
      for i, int_fun in enumerate(integrals):
        _key = arg_names[i][0]
        r = int_fun(params_in[_key], **{arg: params_in[arg] for arg in arg_names[i][1:] if arg in params_in})
        results.append(r)
      return results if isinstance(self.f, joint_eq.JointEq) else results[0]

    return integral_func

  def _build_integrator(self, eq):
    if isinstance(eq, joint_eq.JointEq):
      results = []
      for sub_eq in eq.eqs:
        results.extend(self._build_integrator(sub_eq))
      return results

    else:
      vars, pars, _ = utils.get_args(eq)

      # checking
      if len(vars) != 1:
        raise errors.DiffEqError(f'{self.__class__} only supports numerical integration '
                                 f'for one variable once, while we got {vars} in {eq}. '
                                 f'Please split your multiple variables into multiple '
                                 f'derivative functions.')

      # gradient function
      value_and_grad = math.vector_grad(eq, argnums=0, dyn_vars=self.dyn_var, return_value=True)

      # integration function
      def integral(*args, **kwargs):
        assert len(args) > 0
        dt = kwargs.pop('dt', math.get_dt())
        linear, derivative = value_and_grad(*args, **kwargs)
        phi = math.where(linear == 0., math.ones_like(linear),
                         (math.exp(dt * linear) - 1) / (dt * linear))
        return args[0] + dt * phi * derivative

      return [(integral, vars, pars), ]


register_ode_integrator('exp_euler_auto', ExpEulerAuto)
register_ode_integrator('exp_auto', ExpEulerAuto)



# -*- coding: utf-8 -*-


__all__ = [
  # modules
  'brainobjects', 'layers', 'models',

  # brainobjects
  'DynamicalSystem', 'Container', 'Network',
  'ConstantDelay', 'NeuGroup', 'TwoEndConn',

  # integrators
  'set_default_odeint', 'set_default_sdeint',
  'get_default_odeint', 'get_default_sdeint',

  # monitor
  'Monitor',

  # runners
  'IntegratorRunner', 'DSRunner', 'StructRunner', 'ReportRunner'
]

from . import brainobjects, layers, models
from .brainobjects import *
from .integrators import *
from .monitor import *
from .runners import *



# -*- coding: utf-8 -*-

from typing import Dict

from brainpy import math as bm
from .base import ODEIntegrator

__all__ = [
  'odeint',
  'set_default_odeint',
  'get_default_odeint',
  'register_ode_integrator',
  'get_supported_methods',
]

name2method = {
}

_DEFAULT_DDE_METHOD = 'euler'


def odeint(f=None,
           method='euler',
           var_type=None,
           dt=None,
           name=None,
           # adaptive=None,
           # tol=None,
           show_code=False,
           state_delays: Dict[str, bm.AbstractDelay] = None,
           neutral_delays: Dict[str, bm.NeutralDelay] = None,
           **kwargs):
  """Numerical integration for ODEs.

  Examples
  --------

  .. plot::
    :include-source: True

    >>> import brainpy as bp
    >>> import matplotlib.pyplot as plt
    >>>
    >>> a=0.7;  b=0.8;  tau=12.5;  Vth=1.9
    >>> V = 0;  w = 0  # initial values
    >>>
    >>> @bp.odeint(method='rk4', dt=0.04)
    >>> def integral(V, w, t, Iext):
    >>>   dw = (V + a - b * w) / tau
    >>>   dV = V - V * V * V / 3 - w + Iext
    >>>   return dV, dw
    >>>
    >>> hist_V = []
    >>> for t in bp.math.arange(0, 100, integral.dt):
    >>>     V, w = integral(V, w, t, 0.5)
    >>>     hist_V.append(V)
    >>> plt.plot(bp.math.arange(0, 100, integral.dt), hist_V)
    >>> plt.show()


  Parameters
  ----------
  f : callable, function
    The derivative function.
  method : str
    The shortcut name of the numerical integrator.
  var_type: str
    The type of the variable defined in the equation.
  dt: float
    The numerical integration precision.
  name: str
    The integrator node.
  state_delays: dict
    The state delay variable.
  adaptive: bool
  tol: float
  show_code: bool

  Returns
  -------
  integral : ODEIntegrator
      The numerical solver of `f`.
  """
  method = _DEFAULT_DDE_METHOD if method is None else method
  if method not in name2method:
    raise ValueError(f'Unknown ODE numerical method "{method}". Currently '
                     f'BrainPy only support: {list(name2method.keys())}')

  if f is None:
    return lambda f: name2method[method](f,
                                         var_type=var_type,
                                         dt=dt,
                                         name=name,
                                         show_code=show_code,
                                         state_delays=state_delays,
                                         neutral_delays=neutral_delays,
                                         **kwargs)
  else:
    return name2method[method](f,
                               var_type=var_type,
                               dt=dt,
                               name=name,
                               show_code=show_code,
                               state_delays=state_delays,
                               neutral_delays=neutral_delays,
                               **kwargs)


def set_default_odeint(method):
  """Set the default ODE numerical integrator method for differential equations.

  Parameters
  ----------
  method : str, callable
      Numerical integrator method.
  """
  if not isinstance(method, str):
    raise ValueError(f'Only support string, not {type(method)}.')
  if method not in name2method:
    raise ValueError(f'Unsupported ODE_INT numerical method: {method}.')

  global _DEFAULT_DDE_METHOD
  _DEFAULT_ODE_METHOD = method


def get_default_odeint():
  """Get the default ODE numerical integrator method.

  Returns
  -------
  method : str
      The default numerical integrator method.
  """
  return _DEFAULT_DDE_METHOD


def register_ode_integrator(name, integrator):
  """Register a new ODE integrator.

  Parameters
  ----------
  name: ste
  integrator: type
  """
  if name in name2method:
    raise ValueError(f'"{name}" has been registered in ODE integrators.')
  if not issubclass(integrator, ODEIntegrator):
    raise ValueError(f'"integrator" must be an instance of {ODEIntegrator.__name__}')
  name2method[name] = integrator


def get_supported_methods():
  """Get all supported numerical methods for DDEs."""
  return list(name2method.keys())



# -*- coding: utf-8 -*-

"""
This module provides analysis tools for differential equations.

- The ``symbolic`` module use SymPy symbolic inference to make analysis of
  low-dimensional dynamical system (only sypport ODEs).
- The ``numeric`` module use numerical optimization function to make analysis
  of high-dimensional dynamical system (support ODEs and discrete systems).
- The ``continuation`` module is the analysis package with numerical continuation methods.
- Moreover, we provide several useful functions in ``stability`` module which may
  help your dynamical system analysis.

Details in the following.
"""

from .highdim.slow_points import *

from .lowdim.lowdim_phase_plane import *
from .lowdim.lowdim_bifurcation import *

from . import constants as C
from . import stability
from . import utils



# -*- coding: utf-8 -*-


import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec

__all__ = [
  'get_figure',
]


def get_figure(row_num, col_num, row_len=3, col_len=6):
  """Get the constrained_layout figure.

  Parameters
  ----------
  row_num : int
      The row number of the figure.
  col_num : int
      The column number of the figure.
  row_len : int, float
      The length of each row.
  col_len : int, float
      The length of each column.

  Returns
  -------
  fig_and_gs : tuple
      Figure and GridSpec.
  """
  fig = plt.figure(figsize=(col_num * col_len, row_num * row_len), constrained_layout=True)
  gs = GridSpec(row_num, col_num, figure=fig)
  return fig, gs



# -*- coding: utf-8 -*-

import warnings

from brainpy.dyn import runners as dyn_runner
from brainpy.integrators import runner as intg_runner

__all__ = [
  'IntegratorRunner',
  'DSRunner',
  'StructRunner',
  'ReportRunner'
]


class IntegratorRunner(intg_runner.IntegratorRunner):
  """Integrator runner class.

  .. deprecated:: 2.1.0
     Please use "brainpy.integrators.IntegratorRunner" instead.
  """
  def __init__(self, *args, **kwargs):
    warnings.warn('Please use "brainpy.integrators.IntegratorRunner" instead. '
                  '"brainpy.IntegratorRunner" is deprecated since '
                  'version 2.1.0', DeprecationWarning)
    super(IntegratorRunner, self).__init__(*args, **kwargs)


class DSRunner(dyn_runner.DSRunner):
  """Dynamical system runner class.

  .. deprecated:: 2.1.0
     Please use "brainpy.dyn.DSRunner" instead.
  """
  def __init__(self, *args, **kwargs):
    warnings.warn('Please use "brainpy.dyn.DSRunner" instead. '
                  '"brainpy.DSRunner" is deprecated since '
                  'version 2.1.0', DeprecationWarning)
    super(DSRunner, self).__init__(*args, **kwargs)


class StructRunner(dyn_runner.StructRunner):
  """Dynamical system runner class.

  .. deprecated:: 2.1.0
     Please use "brainpy.dyn.StructRunner" instead.
  """
  def __init__(self, *args, **kwargs):
    warnings.warn('Please use "brainpy.dyn.StructRunner" instead. '
                  '"brainpy.StructRunner" is deprecated since '
                  'version 2.1.0', DeprecationWarning)
    super(StructRunner, self).__init__(*args, **kwargs)


class ReportRunner(dyn_runner.ReportRunner):
  """Dynamical system runner class.

  .. deprecated:: 2.1.0
     Please use "brainpy.dyn.ReportRunner" instead.
  """
  def __init__(self, *args, **kwargs):
    warnings.warn('Please use "brainpy.dyn.ReportRunner" instead. '
                  '"brainpy.ReportRunner" is deprecated since '
                  'version 2.1.0', DeprecationWarning)
    super(ReportRunner, self).__init__(*args, **kwargs)



# -*- coding: utf-8 -*-

import logging
import os
import pickle

import numpy as np

from brainpy import errors
from brainpy.base.collector import TensorCollector

Base = math = None
logger = logging.getLogger('brainpy.base.io')

try:
  import h5py
except (ModuleNotFoundError, ImportError):
  h5py = None

try:
  import scipy.io as sio
except (ModuleNotFoundError, ImportError):
  sio = None

__all__ = [
  'SUPPORTED_FORMATS',
  'save_h5',
  'save_npz',
  'save_pkl',
  'save_mat',
  'load_h5',
  'load_npz',
  'load_pkl',
  'load_mat',
]

SUPPORTED_FORMATS = ['.h5', '.hdf5', '.npz', '.pkl', '.mat']


def _check(module, module_name, ext):
  if module is None:
    raise errors.PackageMissingError(
      '"{package}" must be installed when you want to save/load data with {ext} '
      'format. \nPlease install {package} through "pip install {package}" or '
      '"conda install {package}".'.format(package=module_name, ext=ext)
    )


def _check_missing(variables, filename):
  if len(variables):
    logger.warning(f'There are variable states missed in {filename}. '
                   f'The missed variables are: {list(variables.keys())}.')


def save_h5(filename, all_vars):
  _check(h5py, module_name='h5py', ext=os.path.splitext(filename))
  assert isinstance(all_vars, dict)
  all_vars = TensorCollector(all_vars).unique()

  # save
  f = h5py.File(filename, "w")
  for key, data in all_vars.items():
    f[key] = np.asarray(data.value)
  f.close()


def load_h5(filename, target, verbose=False, check=False):
  global math, Base
  if Base is None: from brainpy.base.base import Base
  if math is None: from brainpy import math
  assert isinstance(target, Base)
  _check(h5py, module_name='h5py', ext=os.path.splitext(filename))

  all_vars = target.vars(method='absolute')
  f = h5py.File(filename, "r")
  for key in f.keys():
    if verbose: print(f'Loading {key} ...')
    var = all_vars.pop(key)
    var[:] = math.asarray(f[key][:])
  f.close()
  if check: _check_missing(all_vars, filename=filename)


def save_npz(filename, all_vars, compressed=False):
  assert isinstance(all_vars, dict)
  all_vars = TensorCollector(all_vars).unique()
  all_vars = {k.replace('.', '--'): np.asarray(v.value) for k, v in all_vars.items()}
  if compressed:
    np.savez_compressed(filename, **all_vars)
  else:
    np.savez(filename, **all_vars)


def load_npz(filename, target, verbose=False, check=False):
  global math, Base
  if Base is None: from brainpy.base.base import Base
  if math is None: from brainpy import math
  assert isinstance(target, Base)

  all_vars = target.vars(method='absolute')
  all_data = np.load(filename)
  for key in all_data.files:
    if verbose: print(f'Loading {key} ...')
    var = all_vars.pop(key)
    var[:] = math.asarray(all_data[key])
  if check: _check_missing(all_vars, filename=filename)


def save_pkl(filename, all_vars):
  assert isinstance(all_vars, dict)
  all_vars = TensorCollector(all_vars).unique()
  targets = {k: np.asarray(v) for k, v in all_vars.items()}
  f = open(filename, 'wb')
  pickle.dump(targets, f, protocol=pickle.HIGHEST_PROTOCOL)
  f.close()


def load_pkl(filename, target, verbose=False, check=False):
  global math, Base
  if Base is None: from brainpy.base.base import Base
  if math is None: from brainpy import math
  assert isinstance(target, Base)
  f = open(filename, 'rb')
  all_data = pickle.load(f)
  f.close()

  all_vars = target.vars(method='absolute')
  for key, data in all_data.items():
    if verbose: print(f'Loading {key} ...')
    var = all_vars.pop(key)
    var[:] = math.asarray(data)
  if check: _check_missing(all_vars, filename=filename)


def save_mat(filename, all_vars):
  assert isinstance(all_vars, dict)
  all_vars = TensorCollector(all_vars).unique()
  _check(sio, module_name='scipy', ext=os.path.splitext(filename))
  all_vars = {k.replace('.', '--'): np.asarray(v.value) for k, v in all_vars.items()}
  sio.savemat(filename, all_vars)


def load_mat(filename, target, verbose=False, check=False):
  global math, Base
  if Base is None: from brainpy.base.base import Base
  if math is None: from brainpy import math
  assert isinstance(target, Base)

  all_data = sio.loadmat(filename)
  all_vars = target.vars(method='absolute')
  for key, data in all_data.items():
    if verbose: print(f'Loading {key} ...')
    var = all_vars.pop(key)
    var[:] = math.asarray(data)
  if check: _check_missing(all_vars, filename=filename)
